import gzip
import logging
import os
import shutil
from ftplib import FTP
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import Optional, Tuple, Any, IO
from urllib.parse import urlparse

import requests

from irrd.conf import get_setting, RPKI_IRR_PSEUDO_SOURCE
from irrd.conf.defaults import DEFAULT_SOURCE_NRTM_PORT
from irrd.rpki.importer import ROADataImporter, ROAParserException
from irrd.rpki.notifications import notify_rpki_invalid_owners
from irrd.rpki.validators import BulkRouteROAValidator
from irrd.scopefilter.validators import ScopeFilterValidator
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.queries import DatabaseStatusQuery
from irrd.utils.whois_client import whois_query
from .parsers import MirrorFileImportParser, NRTMStreamParser

logger = logging.getLogger(__name__)


class RPSLMirrorImportUpdateRunner:
    """
    This RPSLMirrorImportUpdateRunner is the entry point for updating a single
    database mirror, depending on current state.

    If there is no current mirrored data, will call RPSLMirrorFullImportRunner
    to run a new import from full export files. Otherwise, will call
    NRTMImportUpdateStreamRunner to retrieve new updates from NRTM.
    """
    def __init__(self, source: str) -> None:
        self.source = source
        self.full_import_runner = RPSLMirrorFullImportRunner(source)
        self.update_stream_runner = NRTMImportUpdateStreamRunner(source)

    def run(self) -> None:
        self.database_handler = DatabaseHandler()

        try:
            serial_newest_mirror, force_reload = self._status()
            nrtm_enabled = bool(get_setting(f'sources.{self.source}.nrtm_host'))
            logger.debug(f'Most recent mirrored serial for {self.source}: {serial_newest_mirror}, '
                         f'force_reload: {force_reload}, nrtm enabled: {nrtm_enabled}')
            if force_reload or not serial_newest_mirror or not nrtm_enabled:
                self.full_import_runner.run(database_handler=self.database_handler,
                                            serial_newest_mirror=serial_newest_mirror, force_reload=force_reload)
            else:
                self.update_stream_runner.run(serial_newest_mirror, database_handler=self.database_handler)

            self.database_handler.commit()
        except OSError as ose:
            # I/O errors can occur and should not log a full traceback (#177)
            logger.error(f'An error occurred while attempting a mirror update or initial import '
                         f'for {self.source}: {ose}')
        except Exception as exc:
            logger.error(f'An exception occurred while attempting a mirror update or initial import '
                         f'for {self.source}: {exc}', exc_info=exc)
        finally:
            self.database_handler.close()

    def _status(self) -> Tuple[Optional[int], Optional[bool]]:
        query = DatabaseStatusQuery().source(self.source)
        result = self.database_handler.execute_query(query)
        try:
            status = next(result)
            return status['serial_newest_mirror'], status['force_reload']
        except StopIteration:
            return None, None


class FileImportRunnerBase:
    def _retrieve_file(self, url: str, return_contents=True) -> Tuple[str, bool]:
        """
        Retrieve a file from either HTTP(s), FTP or local disk.

        If return_contents is True, the file is read, stripped and then the
        contents are returned. If return_contents is False, the path to a
        local file is returned where the data can be read.

        Return value is a tuple of the contents of the file, or the path to
        the local file, and a boolean to indicate whether the caller should
        unlink the path later.

        If the URL ends in .gz, the file is gunzipped before being processed,
        but only for HTTP(s) and FTP downloads.
        """
        url_parsed = urlparse(url)

        if url_parsed.scheme in ['ftp', 'http', 'https']:
            return self._retrieve_file_download(url, url_parsed, return_contents)
        if url_parsed.scheme == 'file':
            return self._retrieve_file_local(url_parsed.path, return_contents)

        raise ValueError(f'Invalid URL: {url} - scheme {url_parsed.scheme} is not supported')

    def _retrieve_file_download(self, url, url_parsed, return_contents=False) -> Tuple[str, bool]:
        """
        Retrieve a file from HTTP(s) or FTP

        If return_contents is False, the file is read, stripped and then the
        contents are returned. If return_contents is True, the data is written
        to a temporary file, and the path of this file is returned.
        It is the responsibility of the caller to unlink this path later.

        If the URL ends in .gz, the file is gunzipped before being processed,
        but only if return_contents is False.
        """
        destination: IO[Any]
        if return_contents:
            destination = BytesIO()
        else:
            destination = NamedTemporaryFile(delete=False)
        self._download_file(destination, url, url_parsed)
        if return_contents:
            value = destination.getvalue().decode('utf-8').strip()  # type: ignore
            logger.info(f'Downloaded {url}, contained {value}')
            return value, False
        else:
            if url.endswith('.gz'):
                zipped_file = destination
                zipped_file.close()
                destination = NamedTemporaryFile(delete=False)
                logger.debug(f'Downloaded file is expected to be gzipped, gunzipping from {zipped_file.name}')
                with gzip.open(zipped_file.name, 'rb') as f_in:
                    shutil.copyfileobj(f_in, destination)
                os.unlink(zipped_file.name)

            destination.close()

            logger.info(f'Downloaded (and gunzipped if applicable) {url} to {destination.name}')
            return destination.name, True

    def _download_file(self, destination: IO[Any], url: str, url_parsed):
        """
        Download a file from HTTP(s) or FTP.
        The file contents are written to the destination parameter,
        which can be a BytesIO() or a regular file.
        """
        if url_parsed.scheme == 'ftp':
            ftp = FTP(url_parsed.netloc, timeout=600)
            ftp.login()
            ftp.retrbinary(f'RETR {url_parsed.path}', destination.write)
            ftp.quit()
        elif url_parsed.scheme in ['http', 'https']:
            r = requests.get(url, stream=True)
            if r.status_code == 200:
                for chunk in r.iter_content(10240):
                    destination.write(chunk)
            else:
                raise IOError(f'Failed to download {url}: {r.status_code}: {str(r.content)}')

    def _retrieve_file_local(self, path, return_contents=False) -> Tuple[str, bool]:
        if not return_contents:
            if path.endswith('.gz'):
                destination = NamedTemporaryFile(delete=False)
                logger.debug(f'Local file is expected to be gzipped, gunzipping from {path}')
                with gzip.open(path, 'rb') as f_in:
                    shutil.copyfileobj(f_in, destination)
                return destination.name, True
            else:
                return path, False
        with open(path) as fh:
            value = fh.read().strip()
        return value, False


class RPSLMirrorFullImportRunner(FileImportRunnerBase):
    """
    This runner performs a full import from database exports for a single
    mirrored source. URLs for full export file(s), and the URL for the serial
    they match, are provided in configuration.

    Files are downloaded, gunzipped if needed, and then sent through the
    MirrorFileImportParser.
    """
    def __init__(self, source: str) -> None:
        self.source = source

    def run(self, database_handler: DatabaseHandler, serial_newest_mirror: Optional[int]=None, force_reload=False):
        import_sources = get_setting(f'sources.{self.source}.import_source')
        if isinstance(import_sources, str):
            import_sources = [import_sources]
        import_serial_source = get_setting(f'sources.{self.source}.import_serial_source')

        if not import_sources:
            logger.info(f'Skipping full RPSL import for {self.source}, import_source not set.')
            return

        logger.info(f'Running full RPSL import of {self.source} from {import_sources}, serial from {import_serial_source}')

        import_serial = None
        if import_serial_source:
            import_serial = int(self._retrieve_file(import_serial_source, return_contents=True)[0])

            if not force_reload and serial_newest_mirror is not None and import_serial <= serial_newest_mirror:
                logger.info(f'Current newest serial seen from mirror for {self.source} is '
                            f'{serial_newest_mirror}, import_serial is {import_serial}, cancelling import.')
                return

        database_handler.delete_all_rpsl_objects_with_journal(self.source)
        import_data = [self._retrieve_file(import_source, return_contents=False) for import_source in import_sources]

        roa_validator = None
        if get_setting('rpki.roa_source'):
            roa_validator = BulkRouteROAValidator(database_handler)

        database_handler.disable_journaling()
        for import_filename, to_delete in import_data:
            p = MirrorFileImportParser(source=self.source, filename=import_filename, serial=None,
                                       database_handler=database_handler, roa_validator=roa_validator)
            p.run_import()
            if to_delete:
                os.unlink(import_filename)
        if import_serial:
            database_handler.record_serial_newest_mirror(self.source, import_serial)


class ROAImportRunner(FileImportRunnerBase):
    """
    This runner performs a full import of ROA objects.
    The URL file for the ROA export in JSON format is provided
    in the configuration.
    """
    # API consistency with other importers, source is actually ignored
    def __init__(self, source=None):
        pass

    def run(self):
        self.database_handler = DatabaseHandler()

        try:
            self.database_handler.disable_journaling()
            roa_objs = self._import_roas()
            # Do an early commit to make the new ROAs available to other processes.
            self.database_handler.commit()
            # The ROA import does not use journaling, but updating the RPKI
            # status may create journal entries.
            self.database_handler.enable_journaling()

            validator = BulkRouteROAValidator(self.database_handler, roa_objs)
            objs_now_valid, objs_now_invalid, objs_now_not_found = validator.validate_all_routes()
            self.database_handler.update_rpki_status(
                rpsl_objs_now_valid=objs_now_valid,
                rpsl_objs_now_invalid=objs_now_invalid,
                rpsl_objs_now_not_found=objs_now_not_found,
            )
            self.database_handler.commit()
            notified = notify_rpki_invalid_owners(self.database_handler, objs_now_invalid)
            logger.info(f'RPKI status updated for all routes, {len(objs_now_valid)} newly valid, '
                        f'{len(objs_now_invalid)} newly invalid, '
                        f'{len(objs_now_not_found)} newly not_found routes, '
                        f'{notified} emails sent to contacts of newly invalid authoritative objects')

        except OSError as ose:
            # I/O errors can occur and should not log a full traceback (#177)
            logger.error(f'An error occurred while attempting a ROA import: {ose}')
        except ROAParserException as rpe:
            logger.error(f'An exception occurred while attempting a ROA import: {rpe}')
        except Exception as exc:
            logger.error(f'An exception occurred while attempting a ROA import: {exc}', exc_info=exc)
        finally:
            self.database_handler.close()

    def _import_roas(self):
        roa_source = get_setting('rpki.roa_source')
        slurm_source = get_setting('rpki.slurm_source')
        logger.info(f'Running full ROA import from: {roa_source}, SLURM {slurm_source}')

        self.database_handler.delete_all_roa_objects()
        self.database_handler.delete_all_rpsl_objects_with_journal(
            RPKI_IRR_PSEUDO_SOURCE,
            journal_guaranteed_empty=True,
        )

        slurm_data = None
        if slurm_source:
            slurm_data, _ = self._retrieve_file(slurm_source, return_contents=True)

        roa_filename, roa_to_delete = self._retrieve_file(roa_source, return_contents=False)
        with open(roa_filename) as fh:
            roa_importer = ROADataImporter(fh.read(), slurm_data, self.database_handler)
        if roa_to_delete:
            os.unlink(roa_filename)
        logger.info(f'ROA import from {roa_source}, SLURM {slurm_source}, imported {len(roa_importer.roa_objs)} ROAs, running validator')
        return roa_importer.roa_objs


class ScopeFilterUpdateRunner:
    """
    Update the scope filter status for all objects.
    This runner does not actually import anything, the scope filter
    is in the configuration.
    """
    # API consistency with other importers, source is actually ignored
    def __init__(self, source=None):
        pass

    def run(self):
        self.database_handler = DatabaseHandler()

        try:
            validator = ScopeFilterValidator()
            status = validator.validate_all_rpsl_objects(self.database_handler)
            rpsl_objs_now_in_scope, rpsl_objs_now_out_scope_as, rpsl_objs_now_out_scope_prefix = status
            self.database_handler.update_scopefilter_status(
                rpsl_objs_now_in_scope=rpsl_objs_now_in_scope,
                rpsl_objs_now_out_scope_as=rpsl_objs_now_out_scope_as,
                rpsl_objs_now_out_scope_prefix=rpsl_objs_now_out_scope_prefix,
            )
            self.database_handler.commit()
            logger.info(f'Scopefilter status updated for all routes, '
                        f'{len(rpsl_objs_now_in_scope)} newly in scope, '
                        f'{len(rpsl_objs_now_out_scope_as)} newly out of scope AS, '
                        f'{len(rpsl_objs_now_out_scope_prefix)} newly out of scope prefix')

        except Exception as exc:
            logger.error(f'An exception occurred while attempting a scopefilter status update: {exc}', exc_info=exc)
        finally:
            self.database_handler.close()


class NRTMImportUpdateStreamRunner:
    """
    This runner attempts to pull updates from an NRTM stream for a specific
    mirrored database.
    """
    def __init__(self, source: str) -> None:
        self.source = source

    def run(self, serial_newest_mirror: int, database_handler: DatabaseHandler):
        serial_start = serial_newest_mirror + 1
        nrtm_host = get_setting(f'sources.{self.source}.nrtm_host')
        nrtm_port = int(get_setting(f'sources.{self.source}.nrtm_port', DEFAULT_SOURCE_NRTM_PORT))
        if not nrtm_host:
            logger.debug(f'Skipping NRTM updates for {self.source}, nrtm_host not set.')
            return

        end_markings = [
            f'\n%END {self.source}\n',
            f'\n% END {self.source}\n',
            '\n%ERROR',
            '\n% ERROR',
            '\n% Warning: there are no newer updates available',
            '\n% Warning (1): there are no newer updates available',
        ]

        logger.info(f'Retrieving NRTM updates for {self.source} from serial {serial_start} on {nrtm_host}:{nrtm_port}')
        query = f'-g {self.source}:3:{serial_start}-LAST'
        response = whois_query(nrtm_host, nrtm_port, query, end_markings)
        logger.debug(f'Received NRTM response for {self.source}: {response.strip()}')

        stream_parser = NRTMStreamParser(self.source, response, database_handler)
        for operation in stream_parser.operations:
            operation.save(database_handler)



"""Add serial_newest_mirror

Revision ID: 1743f98a456d
Revises: 181670a62643
Create Date: 2020-04-15 20:08:59.925809

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1743f98a456d'
down_revision = '181670a62643'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('database_status', sa.Column('serial_newest_mirror', sa.Integer(), nullable=True))
    op.execute('UPDATE database_status SET serial_newest_mirror = serial_newest_seen')


def downgrade():
    op.drop_column('database_status', 'serial_newest_mirror')



from .access_check import is_client_permitted


class TestIsClientPermitted:
    client_ip = '192.0.2.1'

    def test_no_access_list(self):
        assert is_client_permitted(self.client_ip, 'test.access_list', default_deny=False)
        assert not is_client_permitted(self.client_ip, 'test.access_list', default_deny=True)

    def test_access_list_permitted(self, config_override):
        config_override({
            'test': {
                'access_list': 'test-access-list',
            },
            'access_lists': {
                'test-access-list': ['192.0.2.0/25', '2001:db8::/32'],
            },
        })

        assert is_client_permitted(self.client_ip, 'test.access_list', default_deny=False)
        assert is_client_permitted(self.client_ip, 'test.access_list', default_deny=True)
        assert is_client_permitted(f'::ffff:{self.client_ip}', 'test.access_list', default_deny=True)
        assert is_client_permitted('2001:db8::1', 'test.access_list', default_deny=True)

    def test_access_list_denied(self, config_override):
        config_override({
            'test': {
                'access_list': 'test-access-list',
            },
            'access_lists': {
                'test-access-list': ['192.0.2.128/25', '2001:db8::/32'],
            },
        })

        assert not is_client_permitted(self.client_ip, 'test.access_list', default_deny=False)
        assert not is_client_permitted(f'::ffff:{self.client_ip}', 'test.access_list', default_deny=False)
        assert not is_client_permitted(self.client_ip, 'test.access_list', default_deny=True)

    def test_access_list_denied_invalid_ip(self):
        assert not is_client_permitted('invalid', 'test.access_list', default_deny=False)
        assert not is_client_permitted('invalid', 'test.access_list', default_deny=True)



from collections import defaultdict

import logging
import textwrap
from ordered_set import OrderedSet
from typing import List, Optional, Dict

from irrd.conf import get_setting
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.queries import RPSLDatabaseQuery
from irrd.utils import email
from .parser import parse_change_requests, ChangeRequest
from .parser_state import UpdateRequestStatus, UpdateRequestType
from .validators import ReferenceValidator, AuthValidator

logger = logging.getLogger(__name__)


class ChangeSubmissionHandler:
    """
    The ChangeSubmissionHandler handles the text of one or more requested RPSL changes
    (create, modify or delete), parses, validates and eventually saves
    them. This includes validating references between objects, including
    those part of the same message, and checking authentication.
    """

    def __init__(self, object_texts: str, pgp_fingerprint: str=None, request_meta: Dict[str, Optional[str]]=None) -> None:
        self.database_handler = DatabaseHandler()
        self.request_meta = request_meta if request_meta else {}
        self._pgp_key_id = self._resolve_pgp_key_id(pgp_fingerprint) if pgp_fingerprint else None
        self._handle_object_texts(object_texts)
        self.database_handler.commit()
        self.database_handler.close()

    def _handle_object_texts(self, object_texts: str) -> None:
        reference_validator = ReferenceValidator(self.database_handler)
        auth_validator = AuthValidator(self.database_handler, self._pgp_key_id)
        results = parse_change_requests(object_texts, self.database_handler, auth_validator, reference_validator)

        # When an object references another object, e.g. tech-c referring a person or mntner,
        # an add/update is only valid if those referred objects exist. To complicate matters,
        # the object referred to may be part of this very same submission. For this reason, the
        # reference validator can be provided with all new objects to be added in this submission.
        # However, a possible scenario is that A, B and C are submitted. Object A refers to B,
        # B refers to C, C refers to D and D does not exist - or C fails authentication.
        # At a first scan, A is valid because B exists, B is valid because C exists. C
        # becomes invalid on the first scan, which is why another scan is performed, which
        # will mark B invalid due to the reference to an invalid C, etc. This continues until
        # all references are resolved and repeated scans lead to the same conclusions.
        valid_changes = [r for r in results if r.is_valid()]
        previous_valid_changes: List[ChangeRequest] = []
        loop_count = 0
        loop_max = len(results) + 10

        while valid_changes != previous_valid_changes:
            previous_valid_changes = valid_changes
            reference_validator.preload(valid_changes)
            auth_validator.pre_approve(valid_changes)

            for result in valid_changes:
                result.validate()
            valid_changes = [r for r in results if r.is_valid()]

            loop_count += 1
            if loop_count > loop_max:  # pragma: no cover
                msg = f'Update validity resolver ran an excessive amount of loops, may be stuck, aborting ' \
                      f'processing. Message metadata: {self.request_meta}'
                logger.error(msg)
                raise ValueError(msg)

        for result in results:
            if result.is_valid():
                result.save(self.database_handler)

        self.results = results

    def _resolve_pgp_key_id(self, pgp_fingerprint: str) -> Optional[str]:
        """
        Find a PGP key ID for a given fingerprint.
        This method looks for an actual matching object in the database,
        and then returns the object's PK.
        """
        clean_fingerprint = pgp_fingerprint.replace(' ', '')
        key_id = 'PGPKEY-' + clean_fingerprint[-8:]
        query = RPSLDatabaseQuery().object_classes(['key-cert']).rpsl_pk(key_id)
        results = list(self.database_handler.execute_query(query))

        for result in results:
            if result['parsed_data'].get('fingerpr', '').replace(' ', '') == clean_fingerprint:
                return key_id
        logger.info(f'Message was signed with key {key_id}, but key was not found in the database. Treating message '
                    f'as unsigned. Message metadata: {self.request_meta}')
        return None

    def status(self) -> str:
        """Provide a simple SUCCESS/FAILED string based - former used if all objects were saved."""
        if all([result.status == UpdateRequestStatus.SAVED for result in self.results]):
            return 'SUCCESS'
        return 'FAILED'

    def submitter_report(self) -> str:
        """Produce a human-readable report for the submitter."""
        # flake8: noqa: W293
        successful = [r for r in self.results if r.status == UpdateRequestStatus.SAVED]
        failed = [r for r in self.results if r.status != UpdateRequestStatus.SAVED]
        number_successful_create = len([r for r in successful if r.request_type == UpdateRequestType.CREATE])
        number_successful_modify = len([r for r in successful if r.request_type == UpdateRequestType.MODIFY])
        number_successful_delete = len([r for r in successful if r.request_type == UpdateRequestType.DELETE])
        number_failed_create = len([r for r in failed if r.request_type == UpdateRequestType.CREATE])
        number_failed_modify = len([r for r in failed if r.request_type == UpdateRequestType.MODIFY])
        number_failed_delete = len([r for r in failed if r.request_type == UpdateRequestType.DELETE])

        user_report = self._request_meta_str() + textwrap.dedent(f"""
        SUMMARY OF UPDATE:

        Number of objects found:                  {len(self.results):3}
        Number of objects processed successfully: {len(successful):3}
            Create:      {number_successful_create:3}
            Modify:      {number_successful_modify:3}
            Delete:      {number_successful_delete:3}
        Number of objects processed with errors:  {len(failed):3}
            Create:      {number_failed_create:3}
            Modify:      {number_failed_modify:3}
            Delete:      {number_failed_delete:3}
            
        DETAILED EXPLANATION:
        
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        """)
        for result in self.results:
            user_report += '---\n'
            user_report += result.submitter_report()
            user_report += '\n'
        user_report += '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n'
        return user_report

    def send_notification_target_reports(self):
        # First key is e-mail address of recipient, second is UpdateRequestStatus.SAVED
        # or UpdateRequestStatus.ERROR_AUTH
        reports_per_recipient: Dict[str, Dict[UpdateRequestStatus, OrderedSet]] = defaultdict(dict)
        sources: OrderedSet[str] = OrderedSet()

        for result in self.results:
            for target in result.notification_targets():
                if result.status in [UpdateRequestStatus.SAVED, UpdateRequestStatus.ERROR_AUTH]:
                    if result.status not in reports_per_recipient[target]:
                        reports_per_recipient[target][result.status] = OrderedSet()
                    reports_per_recipient[target][result.status].add(result.notification_target_report())
                    sources.add(result.rpsl_obj_new.source())

        sources_str = '/'.join(sources)
        subject = f'Notification of {sources_str} database changes'
        header = get_setting('email.notification_header', '').format(sources_str=sources_str)
        header += '\nThis message is auto-generated.\n'
        header += 'The request was made by email, with the following details:\n'
        header_saved = textwrap.dedent("""
            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            Some objects in which you are referenced have been created,
            deleted or changed.
            
        """)

        header_failed = textwrap.dedent("""
            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            Some objects in which you are referenced were requested
            to be created, deleted or changed, but *failed* the 
            proper authorisation for any of the referenced maintainers.
            
        """)

        for recipient, reports_per_status in reports_per_recipient.items():
            user_report = header + self._request_meta_str()
            if UpdateRequestStatus.ERROR_AUTH in reports_per_status:
                user_report += header_failed
                for report in reports_per_status[UpdateRequestStatus.ERROR_AUTH]:
                    user_report += f'---\n{report}\n'
                user_report += '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n'
            if UpdateRequestStatus.SAVED in reports_per_status:
                user_report += header_saved
                for report in reports_per_status[UpdateRequestStatus.SAVED]:
                    user_report += f'---\n{report}\n'
                user_report += '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n'

            email.send_email(recipient, subject, user_report)

    def _request_meta_str(self):
        request_meta_str = '\n'.join([f'> {k}: {v}' for k, v in self.request_meta.items() if v])
        if request_meta_str:
            request_meta_str = '\n' + request_meta_str + '\n\n'
        return request_meta_str




# Sample objects used in various tests.
# flake8: noqa: W291,W293

SAMPLE_AS_BLOCK = """as-block:       AS65536 - as065538
descr:          TEST ASN block
remarks:        test remark
mnt-by:         TEST-MNT
changed:        changed@example.com  # only changed line without date
tech-c:         PERSON-TEST
admin-c:        PERSON-TEST
source:         TEST
remarks:        remark
"""

SAMPLE_AS_SET = """as-set:         AS-SETTEST
descr:          description
members:        AS65538, AS65539
members:        AS65537
members:        AS-OTHERSET
tech-c:         PERSON-TEST
admin-c:        PERSON-TEST
notify:         notify@example.com
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_AUT_NUM = """aut-num:        as065537
as-name:        TEST-AS
descr:          description
+               foo
remarks:        ---> Uplinks
export:         to AS3356 announce AS-SETTEST
import:         from AS3356 accept ANY
export:         to AS174 announce AS-SETTEST
import:         from AS174 accept ANY
mp-export:      afi ipv6.unicast to AS174 announce AS-SETTEST
mp-import:      afi ipv6.unicast from AS174 accept ANY
export:         to AS8359 announce AS-SETTEST
import:         from AS8359 accept ANY
export:         to AS3257 announce AS-SETTEST
import:         from AS3257 accept ANY
export:         to AS3549 announce AS-SETTEST
import:         from AS3549 accept ANY
export:         to AS9002 announce AS-SETTEST
import:         from AS9002 accept ANY
mp-export:      afi ipv6.unicast to AS9002 announce AS-SETTEST
mp-import:      afi ipv6.unicast from AS9002 accept ANY
remarks:        ---> Peers
export:         to AS31117 announce AS-SETTEST AS-UAIX
import:         from AS31117 accept AS-ENERGOTEL
export:         to AS8501 announce AS-SETTEST AS-UAIX
import:         from AS8501 accept AS-PLNET
export:         to AS35297 announce AS-SETTEST
import:         from AS35297 accept AS-DL-WORLD
export:         to AS13188 announce AS-SETTEST
import:         from AS13188 accept AS-BANKINFORM
export:         to AS12389 announce AS-SETTEST
import:         from AS12389 accept AS-ROSTELECOM
export:         to AS35395 announce AS-SETTEST
import:         from AS35395 accept AS-GECIXUAIX
export:         to AS50952 announce AS-SETTEST
import:         from AS50952 accept AS-DATAIX
admin-c:        PERSON-TEST
tech-c:         PERSON-TEST
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_DOMAIN = """domain:         2.0.192.in-addr.arpa
admin-c:        PERSON-TEST
tech-c:         PERSON-TEST
zone-c:         PERSON-TEST
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
                # foo
remarks:        remark
"""

SAMPLE_FILTER_SET = """filter-set:     fltr-settest
descr:          Generic anti-bogons filter
remarks:        http://www.iana.org/assignments/ipv4-address-space
filter:         {
                1.0.0.0/8^- ,
                2.0.0.0/8^- ,
                5.0.0.0/8^- ,
                7.0.0.0/8^- ,
                10.0.0.0/8^- ,
                23.0.0.0/8^- ,
                27.0.0.0/8^- ,
                31.0.0.0/8^- ,
                36.0.0.0/8^- ,
                37.0.0.0/8^- ,
                39.0.0.0/8^- ,
                41.0.0.0/8^- ,
                42.0.0.0/8^- ,
                49.0.0.0/8^- ,
                50.0.0.0/8^- ,
                58.0.0.0/8^- ,
                59.0.0.0/8^- ,
                60.0.0.0/8^- ,
                70.0.0.0/8^- ,
                71.0.0.0/8^- ,
                72.0.0.0/8^- ,
                73.0.0.0/8^- ,
                74.0.0.0/8^- ,
                75.0.0.0/8^- ,
                76.0.0.0/8^- ,
                77.0.0.0/8^- ,
                78.0.0.0/8^- ,
                79.0.0.0/8^- ,
                83.0.0.0/8^- ,
                84.0.0.0/8^- ,
                85.0.0.0/8^- ,
                86.0.0.0/8^- ,
                87.0.0.0/8^- ,
                88.0.0.0/8^- ,
                89.0.0.0/8^- ,
                90.0.0.0/8^- ,
                91.0.0.0/8^- ,
                92.0.0.0/8^- ,
                93.0.0.0/8^- ,
                94.0.0.0/8^- ,
                95.0.0.0/8^- ,
                96.0.0.0/8^- ,
                97.0.0.0/8^- ,
                98.0.0.0/8^- ,
                99.0.0.0/8^- ,
                100.0.0.0/8^- ,
                101.0.0.0/8^- ,
                102.0.0.0/8^- ,
                103.0.0.0/8^- ,
                104.0.0.0/8^- ,
                105.0.0.0/8^- ,
                106.0.0.0/8^- ,
                107.0.0.0/8^- ,
                108.0.0.0/8^- ,
                109.0.0.0/8^- ,
                110.0.0.0/8^- ,
                111.0.0.0/8^- ,
                112.0.0.0/8^- ,
                113.0.0.0/8^- ,
                114.0.0.0/8^- ,
                115.0.0.0/8^- ,
                116.0.0.0/8^- ,
                117.0.0.0/8^- ,
                118.0.0.0/8^- ,
                119.0.0.0/8^- ,
                120.0.0.0/8^- ,
                121.0.0.0/8^- ,
                122.0.0.0/8^- ,
                123.0.0.0/8^- ,
                124.0.0.0/8^- ,
                125.0.0.0/8^- ,
                126.0.0.0/8^- ,
                127.0.0.0/8^- ,
                169.254.0.0/16^- ,  # The next two lines were introduced to test blank lines with correct cont. chars
                172.16.0.0/12^- ,   # Note that the trailing whitespace is significant.
\t               
+               
                192.0.2.0/24^- ,
                192.168.0.0/16^- ,
                197.0.0.0/8^- ,
                201.0.0.0/8^- ,
                222.0.0.0/8^- ,
                223.0.0.0/8^- ,
                224.0.0.0/3^-
                }
admin-c:        PERSON-TEST
tech-c:         PERSON-TEST
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_INET_RTR = """inet-rtr:       rtr.example.com
local-as:       AS65537
ifaddr:         192.0.2.1 masklen 30
peer:           BGP4 192.0.2.2 asno(as65530)
admin-c:        PERSON-TEST
tech-c:         PERSON-TEST
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_INET6NUM = """inet6num:       2001:db8::/48
netname:        NET-TEST-V6
country:        DE
admin-c:        PERSON-TEST
tech-c:         PERSON-TEST
mnt-by:         TEST-MNT
status:         ASSIGNED
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_INETNUM = """inetnum:        192.0.2.0 - 192.0.02.255
netname:        NET-TEST-V4
descr:          description
country:        IT
notify:         notify@example.com
admin-c:        PERSON-TEST
tech-c:         PERSON-TEST
status:         ASSIGNED PA
mnt-by:         test-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_KEY_CERT = """key-cert:       PGPKEY-80F238C6
method:         PGP
owner:          Sasha Romijn <sasha@mxsasha.eu>
owner:          Sasha Romijn <sasha@dashcare.nl>
owner:          keybase.io/mxsasha <mxsasha@keybase.io>
owner:          Sasha Romijn <gpg@mxsasha.eu>
fingerpr:       8626 1D8D BEBD A4F5 4692  D64D A838 3BA7 80F2 38C6
certif:         -----BEGIN PGP PUBLIC KEY BLOCK-----
                
                mQINBFnY7YoBEADH5ooPsoR9G/dNxrdHRMJeDHXCSQbwgXWEez5/F8/BZKV9occ/
certif:         jZ7w2wH+Ghj4vTQl1DhuNcxi60qDv9DAPxG73DkBrK0I3fDUJUPrdOKW9SXvZCAq
certif:         LrVEdDVH+YEKhQLlGG7DTODGsfMglL98mn7GD/wD64LtRF3eBAucTIjaOl9hvoqX
certif:         RkH2pQnv/hZKARSS+CYeEpxxdphPM1gm21mSaPYNsKknCIVQsZhHub2PomQWX/AI
certif:         6ZXWmJWObwUlZWEb8CaVBY4+uq6D5sMzUfHsajVJcSHb7l7RnYtCoPaSI5Tj2g22
certif:         q960t9eMEMAtlKm5dvjBcFSQg6sLEJ8YAKa2nxeRyBiapb4kWxn3OhvpDp9XGs3p
certif:         ETyGlT1dUYKcg4QBFKo9x/XydWZRg/tVnZdJX3C8ix+svl4EpIWUowHeSTuQKFK7
certif:         WzS1LzAH6XOaXHRNBLTRYGqb6i/QI6u/hWKqYUu0PtBhu1k0tSmt02ve+YfL3WTZ
certif:         o8VIX2zUSxvnzA4X72Kmc/ni09nNUKlK6MlIAPUxM5cqbGwYRa7WteZCY0iQw+y7
certif:         yC/dNUMT1SrQciFIbd274Z1Mmq7JxxVAmTtQjhFl0dQ/B+PN2m3itm63HsAfiOe0
certif:         Pmj2nprctGShzT4wFrTqi/4NwN8MmEo8n8IHmOye42aLh/FUfso5o5GpPwARAQAB
certif:         tB9TYXNoYSBSb21pam4gPHNhc2hhQG14c2FzaGEuZXU+iQJTBBMBCAA9AhsDBwsJ
certif:         CAcDAgEGFQgCCQoLBBYCAwECHgECF4ACGQEWIQSGJh2Nvr2k9UaS1k2oODungPI4
certif:         xgUCWmh31QAKCRCoODungPI4xkZGD/9+lmMWXj8UgYvjNonad/hcZCjQrWmm+LJf
certif:         DLTS2ZBlNOlkp9rLp2wT+3eTRo2B+e9AlKPfCRTH3e3RRurjqFM5OfWUCLINbPMy
certif:         PfGU9kxCWGunA3mFn2lzAG671vGME7Wjgzga1IsjoIKeRDeFO8qrWBm76ntrXi6n
certif:         q2WB6sEiimrZFdC8DFzwJ4BGwx5GpfL857hlp+2MBCYyJgWOhhQCXVl3r/vcXGq1
certif:         W3qayl8ATASbvihUHLY13BeJIZz2rGKq0VHXAgghu2y42YbyXZji4u3o/Q6hGK/P
certif:         kYFRrJYzJgAVYL4/24xIjEBW381+TYnTzmwKkWYn8Npd5oo/c2ugInh+lr0NspQ5
certif:         vEoNuKgxqC2dS/bD+wH0eSvxQqJdy1eFM5DIxpaNuKc009tUcTBL9zuW4nVQ9fro
certif:         mILFDQeJMK3Ru2ui4hlRqpVTgoKhbnKC10QLjbSOEJSSLWr9kJJrBRbw37Fil/nB
certif:         IisGapf/6zR4C8vLK9Eyo0XdDK3WAimfOE57InuhTimns52R/MZQLnjwmstJmidG
certif:         N0ui29Q48GGbxpc5aXfqV7k5yPJadujKw3SGiRhspG1loQ6SSNacWKJceGS2oIo3
certif:         KVt7/yxkYu2Jn72Ba4FrASiSSmfe4YwySH4q4ru95Es/pp5JQepveg5vSNrOez+i
certif:         1zRcLUpp+4kCHAQTAQgABgUCWdjvjAAKCRCJbs//g9ixtGc+D/4nEFp/G7X/IvZI
certif:         rKiinj9tvSucLskcr3AWFPF7AXXEkc/PK6Ak0sFXUGncKULTy031tSK51mqf7khm
certif:         EUCmhip1h63cyCEbNVMvTyG3GxbKP2Zd1Hpj53qC66cqOfYjjWF1Xjhi2jjAFbaV
certif:         3evE3hdVC0f/1AdlWlueHT6oZHcN9AvXTEKTAalZKYUmjxHRaVsNXxJ/PYwDtC6L
certif:         Pr+tpCcXVqxdeggahse2mKu02CL2SW5YdMst6nsiDHlozk/04wAmwi1kpTy1ZRo9
certif:         V/dXR7qJp1D5SlnfJugAhKzllIx9NtROs1wH35FDnCElyRne1izUaAyEALjVCZ1S
certif:         CziBsNLSGweXDp+rHZf+vAxq7qg3rB1wAuc0i9MiXYuQObxzTPZGXk0BEzVbc83Z
certif:         nIA6LgF5K39z4VUmj37Pd5k/CCA5zQYd+bD74uXJ+5RcyyG7qb3hugiQarhDOWq8
certif:         nCNjTOjXFGke9/kyStszcTNdQ+0/2igr12gp1jWwFQBM5yUjGjaAtAwb4txeZ7U0
certif:         wx/+l8jvzzM7cMPhypmYvLpg/mV5QpImftpAnZ5pYOG/e3f1w6sOvHwYEwVG14yw
certif:         5ZyUOwv8cGCV99ziiKNnUOhp88qcag8Orw2+IdS6ErKZJRlHRMp40jjmoQ25zi5X
certif:         1PCW14mVRiqLmBaVx9n5r2A2YR5vsokCHAQTAQgABgUCWdjvlQAKCRBwjSLshNI3
certif:         AGoaEACkvIgtsA5CZG+CKAorgqcjXgd3J0JaOwoExbDjih9Tay+v1uCFB+AyRYPM
certif:         tOc+s+6gt1hLCUQgpurT/CwKK0gSdKkNAHHPmXa2s6HKreXwo5z6ROdWr456sFO8
certif:         N60QkF0q9G0InnGj+nykS+I3QmNy5W2k9wCyzr//J2FklS2oKMcrjiwWS2Z/+tC5
certif:         Llxmq2H/LDZEVi2PxU91JE8MYcCnLXO4ag2pP6FRyj5DJ8puT74fZ6Vs5mNIH1QB
certif:         /1eQY8XOgg3yyh5B+Tj5yEMJ3+ENxy/7zU2lgI7zGqhvCcqXcIGR/Rv5C14ErGBa
certif:         Ns/VEaQbPchq+4690AAyn+DlhPNBl4Za4SVtNc0a4UrR2x/p0H4uxN/KiLm/OdKu
certif:         61NU/rEsUudUf7Fz+3kMDrBl2dwIRUeNCqq0kr6gCfOxeb9+SIffOw0zJ8OQQrk3
certif:         sPji091smq4x8lOp9h47JSgyFwUN3PXYjrTL0I4KEdXFIViMcuEnLKfxmvXWaRnm
certif:         BndK8urxD+qQOYQOjpFCbuMYR7eBRJK2NSZtFaCIIwsJgfXC0QKwnEWO6DVe7LnK
certif:         20Us5hqA9XGAfoVpdezqmi+W5eesjr10KE0zHp5xMKr11AlEKChIGDWI8iKnEVfo
certif:         GV9Mb8S68wVBZDQA5nqE5zhTdw50N5ESlNGwg/Ln1Sfr/U6dLIkCHAQTAQgABgUC
certif:         WdjvoAAKCRCQW7mjgnYMYDo7EACBmTN1vUaTSWD7t5iyi0L6pu39MvhEv6U8jXxQ
certif:         yO7C2V9GXi4EJhRjmbZAjXOZKrYZ7vKU+ZkjxmAQBk7EGcVJOWl68CT8TNu56HgB
certif:         16idVq3JOQhhckhTFNrFoXE9KSf8nAW6IeTKbt35UjCQ5Dy0Y0q2CW57M5PxGsTW
certif:         3tbpScY6gOjpjcEu70Z0exn6W3GPQEYX1bXyIdnukQQLdPUKijCDBWih9f9X1ZoB
certif:         +L0rloLI3eQmIvojkPe2wVuQ+MdC2Zqdw+umgTSWYeRHawecuVxMKiMUHYlysp0T
certif:         w2Y/6d5nqhEmfZaS5U7C0S7ubGvGQMAA6UuYYYwNC2T0n5Gxvv0K/OwSU1uFlgAn
certif:         zTR9MUFx6qlFLfOvXaHQSd5bOWnGTk64UWvrsPxfrTQhogFtljHz4/Oag6Q7RC0W
certif:         YpgHeBdBEIjckmn7Jw9hYI0P09y/kLS6ecNV9TOL667Vd3bjJ8haSeyWSUfXil+j
certif:         Wt9uXVdojvzl3gXAeYRrJDI89V9owlOXw0bQ+Jod6EqCainY3WPN1AimIBsIomw/
certif:         Ol++RCa1+Wk6h1kRBWIqrIlpQyDLwJWBVed3EhZdO/laVV86jFphOzrkXSXZ2QUH
certif:         xiDju4grylqF3NgWwEvQ3mKxrqVzhlLbZbdP2S1nIGKs0Bw1SMhAj7khRzNAGkpJ
certif:         nd1BY4kCQgQTAQgALAIbAwUJB4YfgAcLCQgHAwIBBhUIAgkKCwQWAgMBAh4BAheA
certif:         BQJZ2O+BAhkBAAoJEKg4O6eA8jjGpyYQAJY2XbzrUJQRw6eQDMu/T2c45bnj4+Ld
certif:         yxowPsbwU7/YwH0871r6xWZBsQ7uQ1MjC1qt2vIutayqywb+Vszg5DtA3xyp97GS
certif:         cjILTd/iURhBYaqNPSUxFMXRl+J4zGwLlglFbdgextGBpx7z9GyfjrKxCflezpOO
certif:         dVc8AR2DS9+3rv/yFs/NNsZXsmhQ3EpwyehpQqH9ksYDwBmuf02KuaZPv1ONCbwF
certif:         t1KHB/JjQrRRJ85jC4uVkAhKY7isGdlh33X7tDi5xGTW2zMf0YCDmqlRmUBbUO8+
certif:         pmkYVWxBIHgF+7HOjwri9AIHWvW0oxSrh01PmiWLG0HgmoJgoumqtG4uJGYUTKAA
certif:         4DDJltd0xBxk8Mu2JQgQ1ZQLKTtxXYYJ5zlU4cC7LeuCFU6ZBV27G8y6yZ5H4HA4
certif:         hPdoxPH4xaPbsp+LXDWl3Ce46x41+IV5qdbTcJabUAynpDPYSKCd6bCeABReAaSU
certif:         Sf0JP3ZEESZGRHODbjiiiZBI6ImGTV8sv/VByJUT49m5hBmVO8e3cxhEoI01voOz
certif:         u1LGPokZ3ansF33vwaO1uEeLViHILFvfAiJ1/7dl5VJzYFe8rLIEHF688igv41eB
certif:         y8CFRjimQj0+tP43rzSxkUuv1aWMtmUkKMupE6349goBV/+XEJuEXHE7UqIiSJz5
certif:         5k+PTeTZhk6htCBTYXNoYSBSb21pam4gPHNhc2hhQGRhc2hjYXJlLm5sPokCUAQT
certif:         AQgAOgIbAwcLCQgHAwIBBhUIAgkKCwQWAgMBAh4BAheAFiEEhiYdjb69pPVGktZN
certif:         qDg7p4DyOMYFAlpod9UACgkQqDg7p4DyOMb9bg/+ORXZnO906haH0SsHxT/wPeba
certif:         zAFDgj27F8fKzoeM8JkqLtXpn+1U4oQIZL6/s61tvpjjdj30BKlqukCltF5R5Zn/
certif:         ItqYaOXteDA9sWIEluQWYZarkZoP9E/3D5OhTEQHhL2WHuSqu+ICUuJiMAFCzcBS
certif:         nEqlfFR5+aOuHBB4TwXjWZPWfWMSC2pSdcND0A8Dsb437dIBcoodLJuRZiKf9muP
certif:         egBDywvhdp9NfiIGpL39jDOeWtOlT5DLmenvmA/CBGQ6Jndz5aKX3njCAeF7YXM9
certif:         fLBpW4en+ZL+kogzulX6csILClLlo6I04FVm8DjogBDRg6cLT/0QcOf/SErDTBfq
certif:         HCy7+5Fh6I4B/YxJ7PltJhrTgR7+m9OjY0VzTJP5faCHu76tOIp4BSRyKIUgVSjU
certif:         TVurGHZUiZL+bKmIAkHIsLFSE+tFTpDxi+tir/o5eskO+gk+HGDnBSjhyqwGM/xC
certif:         T9tlGWDutwnB9kXIDNAp+Cw9l9Tu4T0tq17zNxcFOXsXl+6AhaWbxNtT+P2I5Du/
certif:         XH2p732EJGAGBxA1SZV1FW9swt3bX57mGLlavuP/urgE7phdcv15S1HjYxj/HvJ4
certif:         era0UVn67WMlE61SSIG0f2BS1FXygRCEPhFZmhAdOguIjMLqp5UpsJZvwfPX4I7U
certif:         wSFHhaNtMt+O9MnzL+aJAhwEEwEIAAYFAlnY7kMACgkQiW7P/4PYsbSK0g/+Py5t
certif:         AQwfykWs5LPb6cabXGHALWhhaHFQqf0eeLROsiByrzxY9ZO7n8CCd82JPPFGqpJ5
certif:         +s8auxQbh050Js3w675Mf+5Lnpx9iaa/OF9ku8lT+ktdSpin05Sr0oYsdvERsiP4
certif:         CXRxT4dQ/mUH2lbjGh3xa2FY8DDGMIoygp+Upw3B/fsAODv0g0oK0oKXRu7m3y27
certif:         HXCclR/EoHUqNgGbAChGz8c43z35/eIzh+IQFsdNkLJUHIclDNkZJMg1oMEzdZHM
certif:         J46Wcvs9Schh4uhDSHzlZPVR/8J0eLKvHlAhDqIkc66WjXA8CvFsfphqmU0cjNEE
certif:         ifTljwcYiVbHnlLqAEqwObWElXO8JAFQ37c9IMIjXGTstNGg4xuX6QahJn4JMwnf
certif:         5Xr62A0imPI3uo2r2b4q2Z2sdbMsQucsTdw1939ugAuOYFWYZR5cHrhvOLMxUD3D
certif:         otzjVzsyBoaV2/mdgOLXLyE8ip72XrmEzfSOYV/5BF/+PkPKUoJFiKawSSMKIUQ1
certif:         X8sx02tGHsPGqu3aASaJ7y+n+8aoT7f+oVJuR2SubstOqbi8ZIYrIqtsWmw4Otkw
certif:         hMsRMbNiYGL+ZWHD74++AszC+NVDPf3uO0qfF9uSFO9H5cBSdKG+jm3fJc5tRYnq
certif:         L+qoKUm69Sb9gI7Imhp38mSyIQs0w+JSZwH9OpiJAhwEEwEIAAYFAlnY7lMACgkQ
certif:         kFu5o4J2DGB3MhAAuo2ZjXQQVtgQyJ3rwORwfb8RmIECPHIv1hetgfLYRkJCP4Hz
certif:         WWsyPflKMvJuxbyc5aaDYbWWukqsTZBr+dARYfWmsP7tTNC1QblL5WsWNLwZHF2u
certif:         qWAi5X2kPgbSGb+zVC0KNEfB1yNfw41Fq7XGCK3K9zw/JZxGi8aYAczP7XbkYyGV
certif:         zikQZJTET1mlofzOnqgYoCh7C6/SmJvyY9o726R8Xp8dy7u8CcHkWYLKdoG6aze2
certif:         vi/4U2G25wiJBh2mNToTDaSMEF/8WCyGlxt4SemH1r0rxtpe/fAy/76pXalBSZZC
certif:         mWKxHoQHIvhJWhGqbjiXB9rPvc8d5rLXYzZSYLTq2nAhftygqWyIoS6U1TAs56wL
certif:         yYN0bkV+5yzHVKZ1CVu7LHrs0RKTfXY9JKHF0g+y9wCLLHEM68Rd9BhbIkcdudbN
certif:         T4FjTuKJ16dALHr/c5rLfqeS0dSN7iL/oWulLfs2K44CaGqEFlJmxbRTl/JnHMvq
certif:         igtxKtCzrRZZnHNwfQY6FTA+gn37UeCoNFbGQgIsyPomSOjJwzZRadEWjWg8fC+u
certif:         KOhYEIn3dCd9qXh/YPp7vyCY39IkYPPNbGtb0t6Ycwyuo/chijmcWGlAVXJQgpfk
certif:         sceKR7hKw3KsF2vR9wyjGmwcAwl9eRsKFGotEUG6CAkWPb9ewI94IeiGNSyJAhwE
certif:         EwEIAAYFAlnY7mcACgkQcI0i7ITSNwCBjg//YxHMIO0m1WbSToKwlzryxBWcFf46
certif:         Yq7C+jdeI+mIR3a6P+RJWUMIKIWqcdg/8EgS/lOYp/rsd4Lu5/07N8n2ppJ0BrqN
certif:         VcG0+Wld3f9ALXf4GWQ2JzHfBzGFcQ14gwLQSqTAWDAQw+iPuiqZD98q80NgmbE2
certif:         mgJhrwUiv5mMhARLlvFxZDGyzL2oob6DuNNYyucQLEgcpUbFWeLpgS6xegHXsIat
certif:         2YRkUXazJmeqmANipA4iUSqycYRGVWfQpcNWg0EehFMnLZNu1D3ZUe6/8E83PvqM
certif:         fKPIaIh+fwuEgLVC44tS+OOgstBAzJJ00tV4ESCZf9WjoBTrD6k66i6xONhzV/nv
certif:         YCOK5Z41zZN7l1X5oiq0J8OCDjXEaVi8SCmWOiJHULjMcIRo22BcFyl5W+1lH8zr
certif:         cLRWa43FFt/Ayu7s43KwOs+ILWAc/NxCH/Ri8CcxywMBSRLoWsMFByGNDv7rbpEM
certif:         yAGRQL3M/SDpBeQ2jaOvDpYjJ8vBhrhL8MCPTjXJEHxkY15uAhsW1FR1VYPRI6Yw
certif:         w99XH0ZkYg9GIxt/+hyI+TX/loyGmih/b3OAjEov/Cv34kzjMxewaoRHT3sow6gM
certif:         aJxVGNE1oN+XAYm+9Grsobb0IggV8f4LessNBgTerJn6P/XFfhQR33nbJE3AEn/w
certif:         dR084ZI9I9MlPiiJAj8EEwEIACkFAlnY7d0CGwMFCQeGH4AHCwkIBwMCAQYVCAIJ
certif:         CgsEFgIDAQIeAQIXgAAKCRCoODungPI4xujQD/9ycM82UCOv2kPa5GcHwhVYnK/V
certif:         l0reNViVwjamZ17V4q+H1UU2bXST3RlamLBH7XBwxso2HI3mGSPXs1OkczV1Xq1o
certif:         AH72f1/7lRCtm8VE6ef7ZaGh9yDJnCSG7l8beLoFm1qnSCeaxu28GQY6DP3c1OGZ
certif:         cimQ2v8Qoxx/sY44v8lSqOEbhdTbt2LuhMStWkI1/DiO+CN0FMMULWtXAqlhVvH2
certif:         rItVm/jImqV+TebdbFsq6JTEM8rrrfxBiQ3aV30uj4yKCBXDStiLcfyESGpy+BWh
certif:         nwRal/qI5b4ghAzpBwRED4KSJMW/8IrjMGif01B/iBiGOuEzIewsIJ6RfDuN2BXc
certif:         ovIucU4B7ZAIz4EWYZRNk4TnQBPD6z0bEOwuHmndOrLe1V1x441Kqmb2BvYWENOO
certif:         2euk2sEPo/ZXpNhLcMqijV7d7rJlx7P8NA1eJNz1+0hQBVcpUhGESvUvoVLZS0g+
certif:         KJly5bWJrQ/nEqhfzN1vweWm+houCjpI7cu4xL7uhk6kzaMm4mjXqIcKmOp/wj4v
certif:         sNUWlQinsf+I605FeAnwAZlViucDg8IlXBDq0nETlyyoKytkrnbX5aQtdpZNBglv
certif:         +Yr9tSKmnPV6WZKHDXWcLPPQBaL0dLgCrGlI5ClJPoTNwZSZ3+ibxQLYaMowt1Z/
certif:         sZ1EQUeCFSyHOgef9LQna2V5YmFzZS5pby9teHNhc2hhIDxteHNhc2hhQGtleWJh
certif:         c2UuaW8+iQJQBBMBCAA6AhsDBwsJCAcDAgEGFQgCCQoLBBYCAwECHgECF4AWIQSG
certif:         Jh2Nvr2k9UaS1k2oODungPI4xgUCWmh31QAKCRCoODungPI4xurLD/4tNzm6Y9WX
certif:         APUwsLv8wLokvZlENSvWr2gk157yEeBC7L7GAyq4BsZbCoiMggJMQLcdSaAEwEuS
certif:         hut+sdP8H+pZNXCTgyJ5TGgiUMkxgup+Tw/NlFu5Oe6wxd/XNt2W9NAvRsvYKN5g
certif:         NWsCbT6H9aWXdCXzzujtcczDzANLPHzSZtWoH1oRGoe4014Zl3qGeilAL52BVqaJ
certif:         rMDdw5WOCF7EFQxEWu0yKdBQPajX8Rr5Akq914fXKOBmVGeDhv2EQRUQbeo2gs52
certif:         n/Bwnl6g7YCXhljpejyvK9Pa8uvd6p3yh+WIdz29U7dI5a5MjrttuyFlqknn+ejK
certif:         Bc5AJYB+sWiPPSTwCGXo3gr1VZmlL78d0zmDaPIU5pFDfwNhxjPYKY6MTrqf5mlq
certif:         oHk8Uev1Qwe2Hsdov2JdLmv0Hda7bDsVSQQRp2tDwB23mfxm+jApchdOg7mVbmga
certif:         /V6yCu7/ZFYLYCzJoBToDna3IBElwPMGcNLjASRns6D0UjmJbWMVr+houmrUE/bx
certif:         QTQyAm77+bYJlyESciq8ER8J3uAdCL8+wQC35lzKOwZhF05/xY8qeYGEZfsFOpaZ
certif:         FcBIj4RWJcrX6a99TQkerwjVcRz/qD+wdfnd3aBAeHHUUnGhykWHt1d172Z4X7EW
certif:         VAjOmv1Gby4Kh4xJ8CRcoc6+Cb8lPWNSQIkCHAQTAQgABgUCWdjuQwAKCRCJbs//
certif:         g9ixtPi4EACzdy9asmUCwCw+zRaly9iNRO0BMX32P4za8oo4zEuZzKcnUgV9vGde
certif:         CR3CSfOXrKrITShHC+yOCQ6RAs6uZDZg3IdbiMxCAU/B7hJptOT3ZqrWQK/E6yKT
certif:         bFaFZXArKCBW8eNROLAYodaz4+RcR8ICUB5zfxpTiLVgIKf/uscN8WrxMsdfsNYL
certif:         Qqrd3ubxy9RID41WzaK6ZVgFUVmwP9FJYiDvDU3QRYESe8BcKeYFRAwe59tO+irL
certif:         rld81cnb3/xlifsmnZnuHXDszw0MDzK5/sM6iGktCMgRm4aKgGjLXghoEk3dPFbH
certif:         8NfS/5YQG/pLcsdCwG9GtWEsSUKdYVGvkeTeHHqDSR94ONLe3Yx2GBVplKuuP85I
certif:         DWyIyYrDsKNniSysTPX7sNllCJuS4fIGFuut/nWhi+ampHJ5Oulu6haUdSovMgcq
certif:         dX9uxB92CyyS2VDTo1TRTl4IoJSw2woDX6BRanSMqsqDQE2awHZNjeHqi5IUb0UI
certif:         pHTlAGNS7T4swKWo5P+yQDHcnBvlIsg+6xp4NxWw5QX+WNggzEOLRhsYDKWZ3ckC
certif:         RLGpH5007YGEWTnLCQU/pVn20aFXdvaxi1ZPt0UBPres3y25ObTUn2SfLVqqeCWi
certif:         JtV2MGNT9cbVF31GRVbLBfR+63eUEgwEvpAZ1LO4x5f6tdHwoiNxNYkCHAQTAQgA
certif:         BgUCWdjuUwAKCRCQW7mjgnYMYG2MEADQohyXXnNdZdUVmFt5yCG+hLsxebTWtl0G
certif:         hbPRRl8xO6e+nFRK199Jm1YgNovlWTMp3cLV/LCe5248wg9JjhMTNHtlHCq+l4Ih
certif:         Sm/rEKBY59HZYSVqS5t2ea6vHf0/WpLu0QLDFjGJaIKFBlHRGtwQuzERn1JxU+P6
certif:         j8804V3nPgaG+J0+RZS8YhYRCLz6g2O+1SpXrpMfaVRj1Xnb1Byv6YkWqb1lUdwT
certif:         wmoQKIM8dsykLJHTKv0+UO3bqudZewmY8enfPSzXfZq3YLWp625kRe0g4dzZczHm
certif:         HCoqMqReYmiHBGor8njcT/odmj4lMcr+1tKoybs2JAsz3nfGlsR2hpVdy7t8Vmgp
certif:         GTdsWMorTODV1V3MKGFQS+bky8KJm7mvcba+x+sa8ujIFp5QtP/KixAdCxjeIQDE
certif:         wmWENiZqkXljdxF176ode36HFo8+uZ47Q8TZ29NN2hDYmQe7wfohHV8/5rx6UKPB
certif:         m2JkpRaAZ152yfnK4+pE15xCfoK4831eGK4/ThpR6ZFfp7+V7YUGcg3I8uHV/fDK
certif:         lv+aePJdusWfi2O6NH79YILl3MczuIkgEXYWIUf+rKn3IuV0iQTfZV2bOh3q7LDR
certif:         lQ5w+3joerbZIYSI8JXm1FSdd3zQQma1isGq4Pom/9A5E9z7dRzq0uOAwiH20tOf
certif:         uWLBeQlIDokCHAQTAQgABgUCWdjuZwAKCRBwjSLshNI3AND2EAC7S0zQV/K7Y3oZ
certif:         DjQqUHhSe4kQnoPLHo+JBoX33nmkV7FIJ6rBaxLB4AZvzz3PiwbUc2o7q7g+n0nI
certif:         ejWpALiDhB59oa4ugoiPKcfl1hI1kY+LpbBI8wBrni0GjnlJZQW+7EPY11jyR0MM
certif:         3C8YrQ10EgMLdTuUoLNODmLbnS4U3hatuhkFSwTF1qciBH+DJXM585Ihr3uWkw/L
certif:         BiZz+0DBWfx+uvSOUVKV/tWgNIZkDS84aedVV3ySrmN/0FU66UqfDB7GJjfGoY5V
certif:         fHIzVif2LsypgGvUh3PcQ8gWwjLY6D36yH6oJaIOI1JpkmZgKChFJDfGBjStGg0Y
certif:         rEFira/c4oBUrebKFfipsQVorOVUK2WPP7u7QBB/OadPk1oH3OYWkEKV5wiqNr2A
certif:         bN5f1S94arfn9h+cylSPxDLhfcfAf0JbsH6kYd93q2M8hxhufi77LofgSeqlpKVg
certif:         T0iAuLZa1xfp65MjBhPePXhu4sLZxhBBSWJ1YihiU4g9jkEamGD/jjcI1AGAPiMB
certif:         capmfITMJj+OyPbT7x9XeeiWexKL1+Zo0VTvZ0rIE0OovZLloc9ercErq3e1DKdM
certif:         pOpyvFcps8KnlbvVhSNRYLQgtiI7aOp/yY7WqjwJaTlg3TTQAwfi2V0RGlgKWqSA
certif:         z21EQdrI2JtrKSxsCQLfJn8Gz3odZIkCPwQTAQgAKQUCWdjuNQIbAwUJB4YfgAcL
certif:         CQgHAwIBBhUIAgkKCwQWAgMBAh4BAheAAAoJEKg4O6eA8jjGA5EQALHaW/SJOQde
certif:         vKEzJ3UMTknd5p4uvyi6gEvy6bH9m/3+5CKHz1j1S8gJR/qacOefe+EYo7EElUSC
certif:         vK0Ooi5vJXKd89+SXpcss0g9j5vV+ooEYutDqBGAHI2zdb2cAlLP2aZ2uX/jh2/y
certif:         +bRZinIEDDX87UBwQrWCX0Sn0KWz8erdxkTnZreWRQJOjrG36JuYFTkN4eECvtR2
certif:         Lkfmu2d8tR6qav+psN2/YtQOEJ/pLUSNa3eaTbuaAYtnaTloXuXcWGgQ/hkFQ+R+
certif:         igky12dqSOXMBtQkAOUQrrcW4JlZR2GllI3fAJMAviacRXd32DlGOvtuShsY9Wxl
certif:         XSludWu7nhIhzM9i7ZAN48n1oSIDxkKHYvEaWsFQo2HxN3n1UR56cQk6JIurI5PF
certif:         SOKgv5LWxQZoJBMTAenpEcgFSPbPKw6tQrzmB90XT2rL2LoSMSI6xFEnmeIORtvi
certif:         h42feMahy0fJxW/lKz6EuNlTrtgZnl8wxva4latspM/aWP+cNaZBAh+aaF6kcfHh
certif:         gv2xff0jAC47sgN9k/MaOD5HV+TkJPj9mElzHzFjBGc5hK0I4nT5Ogyx0YhTw7C6
certif:         MpPu5Ia5p675lVQ32JBReAVX+XCp6dC5PVSIDUiHDSuSZQSU7LW5wbc9nPRqrWXU
certif:         f1+YEzT1LM1U3MJJP7K2pJ/g8vopxVc5tB1TYXNoYSBSb21pam4gPGdwZ0BteHNh
certif:         c2hhLmV1PokCUwQTAQgAPQIbAwcLCQgHAwIBBhUIAgkKCwQWAgMBAh4BAheAAhkB
certif:         FiEEhiYdjb69pPVGktZNqDg7p4DyOMYFAlpod9UACgkQqDg7p4DyOMat3RAAgtqq
certif:         kiPg1J5shmWkD9Hc+Xnq6bFoPneDI0WYbJZueodBHU4ww64+KPINuJkJWPZku5hg
certif:         3S7Mns1szWd3KEtSrU+DHPyXsBKQ77Lh8KYaygW8QN3qyfaPk5461vXshYPC4K4G
certif:         uzTTMt3bKg5sfC/Sdkn8W16D2EFU/m2iDuYFhlcVHu6NxViMe7w2CF8uBln65299
certif:         D5INCUSHazht7dpkvViiSXx1cpemaOrOwciHnODgatqM9+X7wa/5st5Vzm+nByxH
certif:         fYttl6dJy2rPhLtd5mDE7/jQm8y590km3F2B0ExMjt2qYhnUEWtbzAm1MyWQy90W
certif:         ni9w/795HBofYHE96+Mr7mnBDJCRpwdxzGqkpX2jiXjpTJEybqFS4jcdP6X2oXFJ
certif:         9aTUO1mugpQ3WCIe4BvOCaof5ueSISkccUGotH6MWGzetkVbLpFQKEJzQFwsSYyE
certif:         2Bck5H3CZEqWeUC5cGHD4K5KeRVe3okMwrsWKEgxKxWuN8Mn7eULsxv0NVRkecRL
certif:         Yds21McnVNKRAs+a42VkG6DR3vXeFE1lgc604wU0RyB2mp1W0w/0/k3Eybgtm42j
certif:         zIEdzhPlylkaaLau8Gbdp1diYMwdqiEP6bMAXUEBVuH3eTp1ihvyWXAd/UiWmuVR
certif:         htK8c83KFzw10k8ljpTiKfLpSu9DphbAWzozJxSJAhwEEwEIAAYFAlnY7aYACgkQ
certif:         iW7P/4PYsbT3Lw//UzAPBV5m3DwtPg6argz03KWq6DFt8jIba4RWJnbJ7K6mNslY
certif:         CcaO4aZiMVmhd9KaeSW67m3XzyvLisLaOF2xSwxdAFkhOZqlnPf/p+EhflN7OVL4
certif:         vdGCULefmTL1j1nGlUbbZz13aevI5t4DXbLY3J5uTBcQBn5N8J1JKlzLNMJ+iAua
certif:         sGjGbs0Qb2R7dIYGB2d+qoX5VlwQIY5IlglPf9RJCzMaOcpBxjRmiUstblrDNUMH
certif:         SdI//143w+b4D5dIJ/ykskmadjaQy7k+B9jiSA62zSN6LVIeb2wOJBxigQ9/D05Q
certif:         cmCbbbaDAw0RMzYU9MMrkJHRO3ppEUDPQWaJ1nogR7JXZOuBuDCKiZfNcpbQXo8j
certif:         ZuBrXol/nCUBMzEJVEuKSzjiu28vZ8jyPL7ZgKc+mm1ZkRj7FyoIx4ZCWvvF30RB
certif:         QkSB9LQ1jLC2V9g2NgnveYPgxpx2J9slb0KeVimw3wj0jOnJHsPhMBJd8JPHNXA8
certif:         RUrtjGYO8Epc7TWBrOcsjmIOZ2PBQ8irwQpK5tDA3ml/AAJ+D6r3Z4/MdMWtzFTV
certif:         7a2u0wm838cABz566YQ3jkkve2lLfcaYpImz4ooiZYl1xvjLMrTu0jzYpChpTBCk
certif:         V6o2wCpjXD38VVHdgGe4XBXLJL/gB/ZZjs1z+uRjucP/tqobyUCvmHwAoN6JAhwE
certif:         EwEIAAYFAlnY7k0ACgkQkFu5o4J2DGDogQ//a+s89OuUUA/ro3qMbP52Igxc+A0f
certif:         a8kPIUE6sCtiS1RQkVbEMcdYG4O+PDakEymiWLgt1xy+/Nx+t8AmfyQ5a5ai9OmH
certif:         9Iwe0CkbqjEX0yUfV8QwlHuPHAaaAqHbqIay6X0zyB0aDWSsKDu1zsV2KTh5kuy5
certif:         IsGbbCQk4EBRqhGA27fQIm0bcjW0MB86BpxCwpDhnNE/2bA1/UY9NXemaeZc1XE1
certif:         QD3q74NANEPjYNbdQxDsXGxD1ctpWvcZW8Vkx5JyKD4bZ7FyoJnw90nv8EJZcdxi
certif:         J2VZDwKvyRP1CiqRI0+XsPcf0iiEWMMGSMDWOIpc8CYvcvUlJK/emPwtMTxJ6yU9
certif:         FKs2P/Oi3WcTnS6xeEDkTucoVyopcT5/bM6eBREdG08RVYpxqNIa1rrwxnWoaun5
certif:         5gqM3ifMibTg12KQ4RIKGKD6ezzwGLhaBHhwOXbJ2ahLDUfCP/we66sZJxRxi6Ef
certif:         Xtwt3fISUDBcbAYpj6vexd+Jqle8YTXe0/h0488kw9s8oBdddH1vsGxFpBJfHmOj
certif:         Sz2pJPaPjZMHMfSBB9COL/fCSiolhbQ8KNatKjv/682SqrZ1ufGDqFb2JQR6/pb9
certif:         78BsVRo3G45KVKZG1Ac526CR3BVRUyJKctCwOO9sW5IUGbpsvlWTA5WmdGBTPiGn
certif:         n5Xby1PIudk5ttKJAhwEEwEIAAYFAlnY7mEACgkQcI0i7ITSNwDNYg//aVAF1gW9
certif:         ZhXyKHvc9kKa4VCthOH+avg+j+YXrzDRP/KIVIokYlNVlg0MutSUbFT37ceUwYTJ
certif:         LuVPxQIe5VYl5kDrl4Ft2nnpoYmumicdbhhstR6PTBjxwQzWcapzc8R/Mvn4xII1
certif:         Rwb0rUDl08/Ezb9H5HWCV1/iusVQRHOsp/klL5QRB9OPAYRL4ZhqSaLiPz9YFVHt
certif:         i2zQl1yjW5iaayfajEIvEWKoZJE9eQ0HX/Uk9mtiTrUO/VZ1IGr8II2Kpb2/vxaL
certif:         0iZQIop7KxPM6rmXzxVymFo3DYQc1LzNs47RcmOMP+Ivsqu2W3QmIrOaSXtBbcLj
certif:         B9SRQGKgelU3qfU5cbQVwiB/R68kU/Xi5V/g3BlH49mshIcUDzU7b0dtxQx7Om2Q
certif:         6+BJCWKo4Q6kzf/ePQr2Heq0GZHpJw5Dv+3yAzAuqQadFFEklq+m3hGWAtZXaSXa
certif:         /c5+IS3+HHfz9o2bc7t/qW7+I6JZC0E6gw/tW8yQZN55J8hVsbHwHBCphxN2IMRC
certif:         u9c4G+14zPrpqB2jAbHrI7iAMhSJa/9OkdA5wuaR1+Hv1Xsx/AVu9h+BWtWrZgjs
certif:         S2p5kwz1MfMZtDhfXjF8y3YPPmQeZl7twxrvYqvFjTq22k5epwYuGPl2JMuV6bwO
certif:         Q2UjMec0MrNbNlq4T3YkGKcTy+4sFIq+rMG5Ag0EWdjtigEQANA0pA0XaEugl04H
certif:         m00lQO0qbSVS6uZXRXsguPiP6LbBLWE/VBg2Dm0rhs5/Ev9wRhOwE1IEKJkn+OqJ
certif:         yJiMebp6rPY+tMplVIKINlaWIMFN9UoMKm72IzY4o/b+YUuWwPMMCgK4Bqd3jzop
certif:         sMVOKn9GX8Pz/hDX+iDQG10IZSzi9fpQZzBk+UmzZqwP+/i4tXxEjfCqnrlwmh0B
certif:         pS1vecaqgRkrTvD3AcTVJVbSekQa4jR292ebRWBIWZm4nATbOf4kVF/DHKuXZIPX
                fXe0acKuNUTfFEDCAM65X7AmVqVx7fcAA9zfei2bWxzDq0wkDE1eEmmfUCvrjdyH
certif:         ensRKKLs8dJWtrB/FB90rsn8FbZUnyXJ9C4BHR4F9zubqj0+B1DQrOl7Hf7hgd6x
certif:         QZpPZ0IxIdECESPRr7yNH9IvMf44N8Zp/zjvD6zwdoIpPiD5ld89a5+xXyYgd7Zr
certif:         MpB5GcRA6cfLETQLXqIoyQWtvtijAbf2oV1xkMxj25LcFRvPBRnF7Qt49U7dsWDt
certif:         AmgXjQnuDvvA0Y4F5NlTiU5kFSOV34I+8l2tHl7Ff0OMcNtdknEX+XPmgTR1x2CE
certif:         4BhFwcBrfFAJghaJ0nCQhPlGTDLraoaBW+9krjeC/YGacUZX28YP044txSlRUzmW
certif:         8+cSao/2yBVljyRq5c85YMk/lK3TABEBAAGJAjYEGAEIACACGwwWIQSGJh2Nvr2k
certif:         9UaS1k2oODungPI4xgUCWpW3SQAKCRCoODungPI4xqLaEACbXTFJEEVnQBeTTTaw
certif:         UQNAyGP41NK/VsxrYbpCs7B6K3HDgTuyTtqsJ3DBR7xl88J3Rj1JYVLfOApalnF5
certif:         6XEwBNBdzQcn10RGo5gIbhvwVMF0RbnSDo9wvwB5Mbdu1F5Mr5mwj9pArGp7z8TY
+               JknUEE/E5krYCZ8HSSSJe5r83uklDR9397+nX/Op0bd/URFJ8ipOxYQPkYJnewxS
certif:         P2pquVrJDBtHPfzXyuVmK4vf2fDa1KX+XSy/JtaUjETebagxBrGQeWdErDHaQtUJ
certif:         U1DS2xGLF2WhirvmmEFDOWKwnmO1HdUSFMGJ1V2LzfkCC1s1OTc3mcYuzzGoM1F7
certif:         TDnLQ9PJqGNqUAHozFuhWfAkX4j5w6k8C3Jb7T7ZhuSswAzEmSfdkg3FHtwQrjXM
certif:         fhbS7tnkmDobMvm7Jo77Lr1CotFBUjUhsR691fA6HATE5svH+CUF0Uq9iPiNIhqK
certif:         4XA9deUxoYxK6InyZS+LI8cKTDjqSzQLBdA65IGmS1K11e5SiK7m7DPO/IH4w/Jd
certif:         M6Ou3r6guvVecNLh42XlKCN2EMew6kKD7GV5MUsvk8w6dlBZAQ12uQadx/BtqhzD
+               NYMaJx7AqP/3hfUjzP1sxnS/HlypR5ULU+H/C0uhTcWbClJ1B1J9MZXUHgjHf2l1
certif:         lxEK/ng5MnwCg1JvzWabWeMR2Q==
certif:         =rpYI
certif:         -----END PGP PUBLIC KEY BLOCK-----
admin-c:        PERSON-TEST
tech-c:         PERSON-TEST
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
"""

SAMPLE_MNTNER = """mntner:         TEST-MNT
admin-c:        PERSON-TEST
notify:         notify@example.net
upd-to:         upd-to@example.net
mnt-nfy:        mnt-nfy@example.net
mnt-nfy:        mnt-nfy2@example.net
auth:           PGPKey-80F238C6
auth:           CRYPT-PW LEuuhsBJNFV0Q  # crypt-password
auth:           MD5-pw $1$fgW84Y9r$kKEn9MUq8PChNKpQhO6BM.  # md5-password
mnt-by:         TEST-MNT
mnt-by:         OTHER1-MNT,OTHER2-MNT
changed:        changed@example.com 20190701 # comment
remarks:        unįcöde tæst 🌈🦄
source:         TEST
remarks:        remark
"""

SAMPLE_PEERING_SET = """peering-set:    prng-settest
descr:          test data for peering-set
peering:        AS65537 at 192.0.2.1
remarks:        MEDIAFAX
tech-c:         PERSON-TEST
tech-c:         DUMY2-TEST
admin-c:        PERSON-TEST
notify:         hostmaster@dnt.ro
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_PERSON = """person:         Test person
address:        DashCare BV
address:        Amsterdam
address:        The Netherlands
phone:          +31 20 000 0000
nic-hdl:        PERSON-TEST
mnt-by:         TEST-MNT
e-mail:         email@example.com
notify:         notify@example.com
changed:        changed@example.com 20190701 # comment
source:         TEST
"""

SAMPLE_ROLE = """role:           DashCare BV
address:        address
phone:          +31200000000
fax-no:         +31200000000
e-mail:         unread@example.com
admin-c:        PERSON-TEST
tech-c:         PERSON-TEST
nic-hdl:        ROLE-TEST
notify:         notify@example.com
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_ROUTE = """route:          192.0.02.0/24
descr:          example route
descr:          the route attribute should have the extra zero removed,
+               but this value should not: 192.0.02.0/24
origin:         AS65537
member-of:      RS-TEST
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
rpki-ov-state: valid  # should be discarded
"""

SAMPLE_ROUTE_SET = """route-set:      RS-TEST
descr:          TEST route set
mbrs-by-ref:    TEST-MNT
tech-c:         PERSON-TEST
admin-c:        PERSON-TEST
mnt-by:         TEST-MNT
members:        RS-OTHER-SET
mp-members:     2001:0dB8::/48
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_ROUTE6 = """route6:         2001:db8::/48
descr:          test route6
origin:         AS65537
mnt-by:         test-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_RTR_SET = """rtr-set:        rtrs-settest
descr:          rtr-set test
members:        rtr.example.com
members:        rtrs-settest
remarks:
tech-c:         PERSON-TEST
admin-c:        PERSON-TEST
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
remarks:        remark
"""

SAMPLE_UNKNOWN_CLASS = """foo-block:          192.0.2.0/24
origin:         AS65537
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
"""

SAMPLE_MALFORMED_EMPTY_LINE = """route:          192.0.2.0/24
origin:         AS65537
mnt-by:         TEST-MNT

changed:        changed@example.com 20190701 # comment
source:         TEST
"""

# https://github.com/irrdnet/irrd/issues/232
SAMPLE_LEGACY_IRRD_ARTIFACT = """*xxte:          192.0.2.0/24
origin:         AS65537
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
"""

SAMPLE_MALFORMED_ATTRIBUTE_NAME = """route:          192.0.2.0/24
origin:         AS65537
$$$-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
"""

SAMPLE_MISSING_MANDATORY_ATTRIBUTE = """route:          192.0.2.0/24
origin:         AS65537
source:         TEST
"""

SAMPLE_UNKNOWN_ATTRIBUTE = """route:          192.0.2.0/24
origin:         AS65537
mnt-by:         TEST-MNT
foo: bar
changed:        changed@example.com 20190701 # comment
source:         TEST
"""

SAMPLE_INVALID_MULTIPLE_ATTRIBUTE = """route:          192.0.2.0/24
origin:         AS65537
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
source:         NOT-TEST
"""

SAMPLE_MALFORMED_PK = """route:          not-a-prefix
origin:         AS65537
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         TEST
"""

SAMPLE_MALFORMED_SOURCE = """route:          192.0.2.0/24
origin:         AS65537
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
source:         +TEST$$$
"""

SAMPLE_MISSING_SOURCE = """route:          192.0.2.0/24
origin:         AS65537
mnt-by:         TEST-MNT
changed:        changed@example.com 20190701 # comment
"""

SAMPLE_LINE_NEITHER_CONTINUATION_NOR_ATTR = """route:          192.0.2.0/24
origin:         AS65537
mnt-by:         TEST-MNT
or
changed:        changed@example.com 20190701 # comment
source:         TEST
"""

KEY_CERT_SIGNED_MESSAGE_VALID = """-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA256

This example was signed with 8626 1D8D BEBD A4F5 4692  D64D A838 3BA7 80F2 38C6.
-----BEGIN PGP SIGNATURE-----

iQIzBAEBCAAdFiEEhiYdjb69pPVGktZNqDg7p4DyOMYFAlrMv68ACgkQqDg7p4Dy
OMbnqhAAnT1H22a+CvSWCjG9Hn7Of+bfr5gpbPNTMlACq4T21rriRWw2EBvERWuH
0FZ02Jwn6tvrw1iXdGuB+BifB7RnB6+B3dfeuKAX0R7D6BdOsJiofv/IopIKAVAC
k7y2t2nRhIG+Q4Fk8/58yVh1Y/axkZZQMhpzUgXhN6eyH+yH1IFkrIc6PUJqfn98
gLkTxhF6Vpy7BhfaHakC4dLVVpE/AfRgf4V2Q6vW5cL4DaRg1uFb29A1Nb9R48kn
X6y8wPcP0rgW73OsEly3buNiogxfPWW+ur0P08do442U1SrnOy/0Vb+cWsvun1R0
AugSbIqneJfkDfrT+hgEGadEsmDbgTTTIGM5EawSCsvG0p0Uzs4gzlPwqebeq/T1
k+9Q4hifIvYFVtZrsjJwuGcs09ps5KYvf6Ps78gT8MssKi7oXS/QRdJUuNFBWWa/
j6nTtsYVerOWL1v1flWkSHrvWQklBOwqqvLHhdG0MFEOx33JTk9kKK+ynysZjMiV
HRJVYrPoCztKU8BA3eAeU1XICWfNPuGXh5LOndgKTzv0urcZkdP84cQmlAe89whD
8bVm0pUkwi2jbzjrXAv6gn+3ecP4ls7pAPJNFPWhuIPuMTmnC3w5AsGGGEzu7EWp
WJUlY+ptExGqyZNT9CJVzVMaZcJEanSnhOsF3G23ObB0Hkz3gLg=
=iSrO
-----END PGP SIGNATURE-----"""

KEY_CERT_SIGNED_MESSAGE_INVALID = """-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA256

This example was signed with 8626 1D8D BEBD A4F5 4692  D64D A838 3BA7 80F2 38C6.
This line was added to invalidate the signature.
-----BEGIN PGP SIGNATURE-----

iQIzBAEBCAAdFiEEhiYdjb69pPVGktZNqDg7p4DyOMYFAlrMv68ACgkQqDg7p4Dy
OMbnqhAAnT1H22a+CvSWCjG9Hn7Of+bfr5gpbPNTMlACq4T21rriRWw2EBvERWuH
0FZ02Jwn6tvrw1iXdGuB+BifB7RnB6+B3dfeuKAX0R7D6BdOsJiofv/IopIKAVAC
k7y2t2nRhIG+Q4Fk8/58yVh1Y/axkZZQMhpzUgXhN6eyH+yH1IFkrIc6PUJqfn98
gLkTxhF6Vpy7BhfaHakC4dLVVpE/AfRgf4V2Q6vW5cL4DaRg1uFb29A1Nb9R48kn
X6y8wPcP0rgW73OsEly3buNiogxfPWW+ur0P08do442U1SrnOy/0Vb+cWsvun1R0
AugSbIqneJfkDfrT+hgEGadEsmDbgTTTIGM5EawSCsvG0p0Uzs4gzlPwqebeq/T1
k+9Q4hifIvYFVtZrsjJwuGcs09ps5KYvf6Ps78gT8MssKi7oXS/QRdJUuNFBWWa/
j6nTtsYVerOWL1v1flWkSHrvWQklBOwqqvLHhdG0MFEOx33JTk9kKK+ynysZjMiV
HRJVYrPoCztKU8BA3eAeU1XICWfNPuGXh5LOndgKTzv0urcZkdP84cQmlAe89whD
8bVm0pUkwi2jbzjrXAv6gn+3ecP4ls7pAPJNFPWhuIPuMTmnC3w5AsGGGEzu7EWp
WJUlY+ptExGqyZNT9CJVzVMaZcJEanSnhOsF3G23ObB0Hkz3gLg=
=iSrO
-----END PGP SIGNATURE-----"""

KEY_CERT_SIGNED_MESSAGE_CORRUPT = """-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA256

This example was signed with 8626 1D8D BEBD A4F5 4692  D64D A838 3BA7 80F2 38C6.
-----BEGIN PGP SIGNATURE-----

iQIzBAEBCAAdFiEEhiYdjb69pPVGktZNqDg7p4DyOMYFAlrMv68ACgkQqDg7p4Dy
0FZ02Jwn6tvrw1iXdGuB+BifB7RnB6+B3dfeuKAX0R7D6BdOsJiofv/IopIKAVAC
8bVm0pUkwi2jbzjrXAv6gn+3ecP4ls7pAPJNFPWhuIPuMTmnC3w5AsGGGEzu7EWp
WJUlY+ptExGqyZNT9CJVzVMaZcJEanSnhOsF3G23ObB0Hkz3gLg
=iSrO
-----END PGP SIGNATURE-----"""

KEY_CERT_SIGNED_MESSAGE_WRONG_KEY = """-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA256

This example was signed with F18F59F232CE9A840D20597B708D22EC84D23700.
-----BEGIN PGP SIGNATURE-----

iQIzBAEBCAAdFiEE8Y9Z8jLOmoQNIFl7cI0i7ITSNwAFAlrN7tkACgkQcI0i7ITS
NwDUWxAAof/us3p+XubKAENmslLPBVqbNjp1uROI3ikWKj469cqNIAjg+HXTnRQe
43KK2VjSxMbgaku531D9iEAilK0tQbOAXmV5FsWFKFZtWh0rxhv9Gu9+jKEkYj8P
YZo8tt3PMKLIejhw26DpTj6Fpq3rUKqsVB+cG/1/P/8zdLEoIUwers9FyAS257Qn
46LCXIwindYv4bkYqqFLMZ1uPVwcfOCuDJ6HE8E3uOytjgi9yKWY1nmjvTA9CwEF
zjeXM6lI4sgctyxmdJyMlh0C5hL0JquB5zeEkGAh03kVxvnPXa1gyrjXhGYCyn5j
dYi8vVI0TBCGr2/jugknJTSU7+rZoySWfF8ZX72AiXrzOXnhc4TkA9nAVwUsoqw5
npHfcc4ilRxUWYQua+dNI2G4d18HKkKWi6d4q9xzdxMPsZBGgaiZ/13o493yocYE
FH/AuTfTK28KJXK9GZpuF787QltjXOc98cijUmvni/6QWfxNAKluYU16/FBGTvV8
6pAcB7E6/95qBdtwomtekOcn3ab+2y8OvZrG3626DB9reGw1wrEkMqb76LjqJ5xS
tVgmazoi0z61pc6BUCdaeSitq3NKzDCYghB9DWBrE3IatIjwWOqPmLYMDkUnp7jl
NW7fkcpU3eaA9cSqZxUaTfUCAYgHNO8wd7sqjWfdMNxVZXTDH48=
=SsN9
-----END PGP SIGNATURE-----"""

SIGNED_PERSON_UPDATE_VALID = """-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA256

person: Test person changed by PGP signed update
address: DashCare BV
address: Amsterdam
address: The Netherlands
phone: +31 20 000 0000
nic-hdl: PERSON-TEST
mnt-by: TEST-MNT
e-mail: email@example.com
changed: changed@example.com 20190701 # comment
source: TEST
-----BEGIN PGP SIGNATURE-----

iQIzBAEBCAAdFiEEhiYdjb69pPVGktZNqDg7p4DyOMYFAl0jNUQACgkQqDg7p4Dy
OMabkBAAq5gCcytpw9UwXMOkLXS2Fd0QfgVRpnc6l0aBmi7uWU6xWe72GU+LaNcC
1vi/OQqT38hXKtDmPMeItPhy5MHanagZZL+ZtVK3SGUaG3rV560Mna2sEnNTeJcb
OvMEg8JXDUP3O4T3kNudTCnBj2d1JhQUNfm7CivWMFe9dxfw+rvirzUnWlhKUZY/
93FGZ1/FpjlAOrLpcFTdvBXajBgpCCHTyTSWBs2KuR2gEWOzIzkyWXIHupwym671
nTg++M/ziPyYJXDv7PqKiBU3DnvSAAOialhk1fse9YW1Dj2dcHz8s2Ex8gv2TcLi
9e6gCF2rOBnusO3yVcqKNBEbpqB+wCbLPGG1C+8n177opTxUipm5kadDBRQy/sZA
P6740cd5Jky1gzWDykch+8ttd8MNVFoNotk1MpauU2zP5/agPuDJuoF6RbCMqX62
MsWp+9c1rlNNUTgQfqxTaEZ+oIj/mLK36iiMQzy0ey9GT/Viuow2WYDjDI5P+OB2
mZp0grLzKK07KdTf/+1WYZb09GhSYzPlKyg12KP3Zoaklh83uYn06mqaeN6YEXYE
zwBzW+p+qqN0rNRMFTNy3WnVVzZY5UWljU83jMBQkXiOSxo72/yIpG89xzi24Bqp
+pewy9PIcK1JBKvGyeO2Gh1K2tsrVYzs7aP5/RmkmUyrQeXa3l4=
=FLEo
-----END PGP SIGNATURE-----
"""

object_sample_mapping = {
    'as-block': SAMPLE_AS_BLOCK,
    'as-set': SAMPLE_AS_SET,
    'aut-num': SAMPLE_AUT_NUM,
    'domain': SAMPLE_DOMAIN,
    'filter-set': SAMPLE_FILTER_SET,
    'inet-rtr': SAMPLE_INET_RTR,
    'inet6num': SAMPLE_INET6NUM,
    'inetnum': SAMPLE_INETNUM,
    'key-cert': SAMPLE_KEY_CERT,
    'mntner': SAMPLE_MNTNER,
    'peering-set': SAMPLE_PEERING_SET,
    'person': SAMPLE_PERSON,
    'role': SAMPLE_ROLE,
    'route': SAMPLE_ROUTE,
    'route-set': SAMPLE_ROUTE_SET,
    'route6': SAMPLE_ROUTE6,
    'rtr-set': SAMPLE_RTR_SET,
}

TEMPLATE_ROUTE_OBJECT = """route:          [mandatory]  [single]    [primary/look-up key]
descr:          [optional]   [multiple]  []
origin:         [mandatory]  [single]    [primary key]
holes:          [optional]   [multiple]  []
member-of:      [optional]   [multiple]  [look-up key, weak references route-set]
inject:         [optional]   [multiple]  []
aggr-bndry:     [optional]   [single]    []
aggr-mtd:       [optional]   [single]    []
export-comps:   [optional]   [single]    []
components:     [optional]   [single]    []
admin-c:        [optional]   [multiple]  [look-up key, strong references role/person]
tech-c:         [optional]   [multiple]  [look-up key, strong references role/person]
geoidx:         [optional]   [multiple]  []
roa-uri:        [optional]   [single]    []
remarks:        [optional]   [multiple]  []
notify:         [optional]   [multiple]  []
mnt-by:         [mandatory]  [multiple]  [look-up key, strong references mntner]
changed:        [optional]   [multiple]  []
source:         [mandatory]  [single]    []
"""

TEMPLATE_PERSON_OBJECT = """person:    [mandatory]  [single]    [look-up key]
address:   [mandatory]  [multiple]  []
phone:     [mandatory]  [multiple]  []
fax-no:    [optional]   [multiple]  []
e-mail:    [mandatory]  [multiple]  []
nic-hdl:   [mandatory]  [single]    [primary/look-up key]
remarks:   [optional]   [multiple]  []
notify:    [optional]   [multiple]  []
mnt-by:    [mandatory]  [multiple]  [look-up key, strong references mntner]
changed:   [optional]   [multiple]  []
source:    [mandatory]  [single]    []
"""



import os
import textwrap

import pytest
import signal
import yaml
from typing import Dict

from . import get_setting, ConfigurationError, config_init, is_config_initialised, get_configuration


@pytest.fixture()
def save_yaml_config(tmpdir, monkeypatch):
    def _save(config: Dict, run_init=True):
        tmp_file = tmpdir + '/config.yaml'
        with open(tmp_file, 'w') as fh:
            fh.write(yaml.safe_dump(config))
        if run_init:
            config_init(str(tmp_file))
    return _save


class TestConfiguration:
    def test_file_not_existing(self, monkeypatch, tmpdir):
        with pytest.raises(ConfigurationError) as ce:
            config_init(str(tmpdir + '/doesnotexist.yaml'))
        assert 'Error opening config file' in str(ce.value)

    def test_load_invalid_yaml(self, monkeypatch, tmpdir):
        tmp_file = tmpdir + '/config.yaml'
        fh = open(tmp_file, 'w')
        fh.write('  >foo')
        fh.close()

        with pytest.raises(ConfigurationError) as ce:
            config_init(str(tmp_file))
        assert 'Error parsing YAML file' in str(ce.value)

    def test_load_string_file(self, save_yaml_config):
        with pytest.raises(ConfigurationError) as ce:
            save_yaml_config('foo')
        assert 'Could not find root item "irrd" in config file' in str(ce.value)

    def test_load_empty_config(self, save_yaml_config):
        with pytest.raises(ConfigurationError) as ce:
            save_yaml_config({})
        assert 'Could not find root item "irrd" in config file' in str(ce.value)

    def test_load_valid_reload_valid_config(self, monkeypatch, save_yaml_config, tmpdir, caplog):
        logfile = str(tmpdir + '/logfile.txt')
        config = {
            'irrd': {
                'database_url': 'db-url',
                'redis_url': 'redis-url',
                'piddir': str(tmpdir),
                'email': {
                    'from': 'example@example.com',
                    'smtp': '192.0.2.1'
                },
                'rpki': {
                    'roa_source': None,
                },
                'scopefilter': {
                    'prefixes': ['10/8'],
                    'asns': ['23456', '10-20']
                },
                'access_lists': {
                    'valid-list': {
                        '192/24',
                        '192.0.2.1',
                        '2001:db8::32',
                        '2001:db8::1',
                    }
                },
                'auth': {
                    'gnupg_keyring': str(tmpdir)
                },
                'sources_default': ['TESTDB2', 'TESTDB'],
                'sources': {
                    'TESTDB': {
                        'authoritative': True,
                        'keep_journal': True,
                    },
                    'TESTDB2': {
                        'nrtm_host': '192.0.2.1',
                        'nrtm_port': 43,
                        'import_serial_source': 'ftp://example.com/serial',
                        'keep_journal': True,
                    },
                    'TESTDB3': {
                        'export_destination_unfiltered': '/tmp',
                        'nrtm_access_list_unfiltered': 'valid-list',
                    },
                    # RPKI source permitted, rpki.roa_source not set
                    'RPKI': {},
                },
                'log': {
                    'level': 'DEBUG',
                    'logfile_path': logfile
                },

            }
        }
        save_yaml_config(config)
        assert is_config_initialised()

        config['irrd']['sources_default'] = ['TESTDB2']
        save_yaml_config(config, run_init=False)

        # Unchanged, no reload performed
        assert list(get_setting('sources_default')) == ['TESTDB2', 'TESTDB']

        os.kill(os.getpid(), signal.SIGHUP)
        assert list(get_setting('sources_default')) == ['TESTDB2']

        logfile_contents = open(logfile).read()
        assert 'Configuration successfully (re)loaded from ' in logfile_contents

    def test_load_custom_logging_config(self, monkeypatch, save_yaml_config, tmpdir, caplog):
        logfile = str(tmpdir + '/logfile.txt')
        logging_config_path = str(tmpdir + '/logging.py')
        with open(logging_config_path, 'w') as fh:
            fh.write(textwrap.dedent("""
                LOGGING = {
                    'version': 1,
                    'disable_existing_loggers': False,
                    'formatters': {
                        'verbose': {
                            'format': '%(asctime)s irrd[%(process)d]: [%(name)s#%(levelname)s] %(message)s'
                        },
                    },
                    'handlers': {
                        'file': {
                            'class': 'logging.handlers.WatchedFileHandler',
                            'filename': '""" + logfile + """',
                            'formatter': 'verbose',
                        },
                    },
                    'loggers': {
                        '': {
                            'handlers': ['file'],
                            'level': 'DEBUG',
                        },
                    }
                }
            """))
        config = {
            'irrd': {
                'database_url': 'db-url',
                'redis_url': 'redis-url',
                'piddir': str(tmpdir),
                'email': {
                    'from': 'example@example.com',
                    'smtp': '192.0.2.1'
                },
                'rpki': {
                    'roa_source': None,
                },
                'auth': {
                    'gnupg_keyring': str(tmpdir)
                },
                'log': {
                    'logging_config_path': logging_config_path,
                },

            }
        }
        save_yaml_config(config)
        assert is_config_initialised()
        assert get_configuration().logging_config['handlers']['file']['filename'] == logfile

    def test_load_valid_reload_invalid_config(self, save_yaml_config, tmpdir, caplog):
        save_yaml_config({
            'irrd': {
                'database_url': 'db-url',
                'redis_url': 'redis-url',
                'piddir': str(tmpdir),
                'email': {
                    'from': 'example@example.com',
                    'smtp': '192.0.2.1'
                },
                'access-lists': {
                    'valid-list': {
                        '192/24',
                        '192.0.2.1',
                        '2001:db8::32',
                        '2001:db8::1',
                    }
                },
                'auth': {
                    'gnupg_keyring': str(tmpdir)
                },
                'rpki': {
                    'roa_source': 'https://example.com/roa.json',
                    'notify_invalid_enabled': False,
                },
                'sources_default': ['TESTDB2', 'TESTDB', 'RPKI'],
                'sources': {
                    'TESTDB': {
                        'authoritative': True,
                        'keep_journal': True,
                    },
                    'TESTDB2': {
                        'nrtm_host': '192.0.2.1',
                        'nrtm_port': 43,
                        'import_serial_source': 'ftp://example.com/serial',
                        'keep_journal': True,
                        'import_timer': '1234',
                    },
                },

            }
        })

        save_yaml_config({}, run_init=False)
        os.kill(os.getpid(), signal.SIGHUP)
        assert list(get_setting('sources_default')) == ['TESTDB2', 'TESTDB', 'RPKI']
        assert 'Errors found in configuration, continuing with current settings' in caplog.text
        assert 'Could not find root item "irrd"' in caplog.text

    def test_load_invalid_config(self, save_yaml_config, tmpdir):
        config = {
            'irrd': {
                'database_readonly': True,
                'piddir': str(tmpdir + '/does-not-exist'),
                'server': {
                    'whois': {
                        'access_list': 'doesnotexist',
                    },
                    'http': {
                        'access_list': ['foo'],
                    },
                },
                'email': {
                    'footer': {'a': 1},
                    'recipient_override': 'invalid-mail',
                },
                'access_lists': {
                    'bad-list': {
                        '192.0.2.2.1'
                    },
                },
                'rpki': {
                    'roa_source': 'https://example.com/roa.json',
                    'roa_import_timer': 'foo',
                    'notify_invalid_subject': [],
                    'notify_invalid_header': [],
                },
                'scopefilter': {
                    'prefixes': ['invalid-prefix'],
                    'asns': ['invalid', '10-invalid'],
                },
                'sources_default': ['DOESNOTEXIST-DB'],
                'sources': {
                    'TESTDB': {
                        'keep_journal': True,
                        'import_timer': 'foo',
                        'export_timer': 'bar',
                        'nrtm_host': '192.0.2.1',
                    },
                    'TESTDB2': {
                        'authoritative': True,
                        'nrtm_host': '192.0.2.1',
                        'nrtm_port': 'not a number',
                        'nrtm_access_list': 'invalid-list',
                    },
                    'TESTDB3': {
                        'authoritative': True,
                        'import_source': '192.0.2.1',
                        'nrtm_access_list_unfiltered': 'invalid-list',
                    },
                    # Not permitted, rpki.roa_source is set
                    'RPKI': {},
                    'lowercase': {},
                    'invalid char': {},
                },
                'log': {
                    'level': 'INVALID',
                    'logging_config_path': 'path',
                }
            }
        }

        with pytest.raises(ConfigurationError) as ce:
            save_yaml_config(config)

        assert 'Setting database_url is required.' in str(ce.value)
        assert 'Setting redis_url is required.' in str(ce.value)
        assert 'Setting piddir is required and must point to an existing directory.' in str(ce.value)
        assert 'Setting email.from is required and must be an email address.' in str(ce.value)
        assert 'Setting email.smtp is required.' in str(ce.value)
        assert 'Setting email.footer must be a string, if defined.' in str(ce.value)
        assert 'Setting email.recipient_override must be an email address if set.' in str(ce.value)
        assert 'Setting auth.gnupg_keyring is required.' in str(ce.value)
        assert 'Access lists doesnotexist, invalid-list referenced in settings, but not defined.' in str(ce.value)
        assert 'Setting server.http.access_list must be a string, if defined.' in str(ce.value)
        assert 'Invalid item in access list bad-list: IPv4 Address with more than 4 bytes.' in str(ce.value)
        assert 'Invalid item in prefix scopefilter: invalid-prefix' in str(ce.value)
        assert 'Invalid item in asn scopefilter: invalid.' in str(ce.value)
        assert 'Invalid item in asn scopefilter: 10-invalid.' in str(ce.value)
        assert 'Setting sources contains reserved source name: RPKI' in str(ce.value)
        assert 'Setting keep_journal for source TESTDB can not be enabled unless either ' in str(ce.value)
        assert 'Setting nrtm_host for source TESTDB can not be enabled without setting import_serial_source.' in str(ce.value)
        assert 'Setting authoritative for source TESTDB2 can not be enabled when either nrtm_host or import_source are set.' in str(ce.value)
        assert 'Setting authoritative for source TESTDB3 can not be enabled when either nrtm_host or import_source are set.' in str(ce.value)
        assert 'Source TESTDB can not have authoritative, import_source or nrtm_host set when database_readonly is enabled.' in str(ce.value)
        assert 'Source TESTDB3 can not have authoritative, import_source or nrtm_host set when database_readonly is enabled.' in str(ce.value)
        assert 'Setting nrtm_port for source TESTDB2 must be a number.' in str(ce.value)
        assert 'Setting rpki.roa_import_timer must be set to a number.' in str(ce.value)
        assert 'Setting rpki.notify_invalid_subject must be a string, if defined.' in str(ce.value)
        assert 'Setting rpki.notify_invalid_header must be a string, if defined.' in str(ce.value)
        assert 'Setting import_timer for source TESTDB must be a number.' in str(ce.value)
        assert 'Setting export_timer for source TESTDB must be a number.' in str(ce.value)
        assert 'Invalid source name: lowercase' in str(ce.value)
        assert 'Invalid source name: invalid char' in str(ce.value)
        assert 'but rpki.notify_invalid_enabled is not set' in str(ce.value)
        assert 'Setting sources_default contains unknown sources: DOESNOTEXIST-DB' in str(ce.value)
        assert 'Invalid log.level: INVALID' in str(ce.value)
        assert 'Setting log.logging_config_path can not be combined' in str(ce.value)


class TestGetSetting:
    setting_name = 'server.whois.interface'
    env_name = 'IRRD_SERVER_WHOIS_INTERFACE'

    def test_get_setting_default(self, monkeypatch):
        monkeypatch.delenv(self.env_name, raising=False)
        assert get_setting(self.setting_name) == '::0'

    def test_get_setting_env(self, monkeypatch):
        monkeypatch.setenv(self.env_name, 'env_value')
        assert get_setting(self.setting_name) == 'env_value'



#!/usr/bin/env python
# flake8: noqa: E402
"""
This is a helper script to run RPSL data through the parser and, optionally,
insert it into the database.
"""
import argparse
import sys

from pathlib import Path
from typing import Set

from irrd.storage.models import JournalEntryOrigin

sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.conf import CONFIG_PATH_DEFAULT, config_init, get_setting
from irrd.storage.database_handler import DatabaseHandler
from irrd.rpsl.parser import UnknownRPSLObjectClassException
from irrd.rpsl.rpsl_objects import rpsl_object_from_text
from irrd.utils.text import split_paragraphs_rpsl


class RPSLParse:
    obj_parsed = 0
    obj_errors = 0
    obj_unknown = 0
    unknown_object_classes: Set[str] = set()
    database_handler = None

    def main(self, filename, strict_validation, database, show_info=True):
        self.show_info = show_info
        if database:
            self.database_handler = DatabaseHandler()
            self.database_handler.disable_journaling()

        if filename == '-':  # pragma: no cover
            f = sys.stdin
        else:
            f = open(filename, encoding='utf-8', errors='backslashreplace')

        for paragraph in split_paragraphs_rpsl(f):
            self.parse_object(paragraph, strict_validation)

        print(f'Processed {self.obj_parsed} objects, {self.obj_errors} with errors')
        if self.obj_unknown:
            unknown_formatted = ', '.join(self.unknown_object_classes)
            print(f'Ignored {self.obj_unknown} objects due to unknown object classes: {unknown_formatted}')

        if self.database_handler:
            self.database_handler.commit()
            self.database_handler.close()

    def parse_object(self, rpsl_text, strict_validation):
        try:
            self.obj_parsed += 1
            obj = rpsl_object_from_text(rpsl_text.strip(), strict_validation=strict_validation)
            if (obj.messages.messages() and self.show_info) or obj.messages.errors():
                if obj.messages.errors():
                    self.obj_errors += 1

                print(rpsl_text.strip())
                print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
                print(obj.messages)
                print('\n=======================================\n')

            if self.database_handler and obj and not obj.messages.errors():
                self.database_handler.upsert_rpsl_object(obj, JournalEntryOrigin.mirror)

        except UnknownRPSLObjectClassException as e:
            self.obj_unknown += 1
            self.unknown_object_classes.add(str(e).split(':')[1].strip())
        except Exception as e:  # pragma: no cover
            print('=======================================')
            print(rpsl_text)
            print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
            raise e


def main():  # pragma: no cover
    description = """Run RPSL data through the IRRD processor. For each object that resulted in messages emitted by
                     the parser, the object is printed followed by the messages. Optionally, insert objects into
                     the database."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    parser.add_argument('--hide-info', dest='hide_info', action='store_true',
                        help='hide INFO messages')
    parser.add_argument('--strict', dest='strict_validation', action='store_true',
                        help='use strict validation (errors on e.g. unknown or missing attributes)')
    parser.add_argument('--database-destructive-overwrite', dest='database', action='store_true',
                        help='insert all valid objects into the IRRD database - OVERWRITING ANY EXISTING ENTRIES, if '
                             'they have the same RPSL primary key and source')
    parser.add_argument('input_file', type=str,
                        help='the name of a file to read, or - for stdin')
    args = parser.parse_args()

    config_init(args.config_file_path)
    if get_setting('database_readonly'):
        print('Unable to run, because database_readonly is set')
        sys.exit(-1)
        
    RPSLParse().main(args.input_file, args.strict_validation, args.database, not args.hide_info)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import setuptools
from irrd import __version__

with open('README.rst', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='irrd',
    version=__version__,
    author='DashCare for NTT Ltd.',
    author_email='irrd@dashcare.nl',
    description='Internet Routing Registry daemon (IRRd)',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/irrdnet/irrd',
    packages=setuptools.find_packages(
        exclude=['*.tests', '*.tests.*', 'tests.*', 'tests', 'irrd.integration_tests']
    ),
    python_requires='>=3.6',
    package_data={'': ['*.txt', '*.yaml', '*.mako']},
    install_requires=[
        # This list must be kept in sync with requirements.txt version-wise,
        # but should not include packages used for testing, generating docs
        # or packages.
        'python-gnupg==0.4.6',
        'passlib==1.7.2',
        'IPy==1.00',
        'ordered-set==4.0.2',
        'dotted==0.1.8',
        'beautifultable==0.8.0',
        'PyYAML==5.3.1',
        'datrie==0.8.2',
        'setproctitle==1.1.10',
        'python-daemon==2.2.4',
        'pid==3.0.4',
        'redis==3.5.3',
        'hiredis==1.1.0',
        'requests==2.24.0',
        'pytz==2020.1',
        'psycopg2-binary==2.8.5',
        'SQLAlchemy==1.3.18',
        'alembic==1.4.2',
        'ujson==3.0.0',
    ],
    extras_require={
        ':python_version < "3.7"': [
            'dataclasses==0.7',
        ],
    },
    entry_points={
        'console_scripts': [
            'irrd = irrd.daemon.main:main',
            'irrd_submit_email = irrd.scripts.submit_email:main',
            'irrd_database_upgrade = irrd.scripts.database_upgrade:main',
            'irrd_database_downgrade = irrd.scripts.database_downgrade:main',
            'irrd_load_database = irrd.scripts.load_database:main',
            'irrd_update_database = irrd.scripts.update_database:main',
            'irrd_set_last_modified_auth = irrd.scripts.set_last_modified_auth:main',
            'irrd_mirror_force_reload = irrd.scripts.mirror_force_reload:main',
        ],
    },
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: OS Independent',
    ],
)



#!/usr/bin/env python
# flake8: noqa: E402
import argparse
import logging
import sys

from pathlib import Path

"""
Load an RPSL file into the database.
"""

logger = logging.getLogger(__name__)
sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.conf import config_init, CONFIG_PATH_DEFAULT
from irrd.storage.queries import RPSLDatabaseQuery
from irrd.storage.database_handler import DatabaseHandler
from irrd.rpsl.rpsl_objects import rpsl_object_from_text

def load_pgp_keys(source: str) -> None:
    dh = DatabaseHandler()
    query = RPSLDatabaseQuery(column_names=['rpsl_pk', 'object_text'])
    query = query.sources([source]).object_classes(['key-cert'])
    keycerts = dh.execute_query(query)

    for keycert in keycerts:
        rpsl_pk = keycert["rpsl_pk"]
        print(f'Loading key-cert {rpsl_pk}')
        # Parsing the keycert in strict mode will load it into the GPG keychain
        result = rpsl_object_from_text(keycert['object_text'], strict_validation=True)
        if result.messages.errors():
            print(f'Errors in PGP key {rpsl_pk}: {result.messages.errors()}')

    print('All valid key-certs loaded into the GnuPG keychain.')
    dh.close()


def main():  # pragma: no cover
    description = """Load all PGP keys from key-cert objects for a specific source into the GnuPG keychain."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--config', dest='config_file_path', type=str,
                        help=f'use a different IRRd config file (default: {CONFIG_PATH_DEFAULT})')
    parser.add_argument('source', type=str,
                        help='the name of the source for which to load PGP keys')
    args = parser.parse_args()

    config_init(args.config_file_path)

    load_pgp_keys(args.source)


if __name__ == '__main__':  # pragma: no cover
    main()



#!/usr/bin/env python
# flake8: noqa: E402
"""
This script will run a list of queries against two IRRD instances,
and report significant results.
"""

import argparse
import difflib
import sys

import re
from IPy import IP
from ordered_set import OrderedSet
from pathlib import Path
from typing import Optional


sys.path.append(str(Path(__file__).resolve().parents[2]))

from irrd.rpsl.rpsl_objects import rpsl_object_from_text
from irrd.utils.text import splitline_unicodesafe, split_paragraphs_rpsl
from irrd.utils.whois_client import whois_query_irrd, whois_query, WhoisQueryError


SSP_QUERIES = ['!6', '!g', '!i']
ASDOT_RE = re.compile(r'as\d+\.\d*', flags=re.IGNORECASE)


class QueryComparison:
    queries_run = 0
    queries_different = 0
    queries_both_error = 0
    queries_invalid = 0
    queries_mirror = 0

    def __init__(self, input_file, host_reference, port_reference, host_tested, port_tested):
        self.host_reference = host_reference
        self.port_reference = port_reference
        self.host_tested = host_tested
        self.port_tested = port_tested

        if input_file == '-':
            f = sys.stdin
        else:
            f = open(input_file, encoding='utf-8', errors='backslashreplace')

        for query in f.readlines():
            query = query.strip() + '\n'
            if query == '!!\n':
                continue
            self.queries_run += 1
            error_reference = None
            error_tested = None
            response_reference = None
            response_tested = None

            # ignore version or singular source queries
            if query.lower().startswith('!v') or query.lower().startswith('!s'):
                continue

            if (query.startswith('-x') and not query.startswith('-x ')) or re.search(ASDOT_RE, query):
                self.queries_invalid += 1
                continue

            # ignore queries asking for NRTM data or mirror serial status
            if query.lower().startswith('-g ') or query.lower().startswith('!j'):
                self.queries_mirror += 1
                continue

            if query.startswith('!'):  # IRRD style query
                try:
                    response_reference = whois_query_irrd(self.host_reference, self.port_reference, query)
                except ConnectionError as ce:
                    error_reference = str(ce)
                except WhoisQueryError as wqe:
                    error_reference = str(wqe)
                except ValueError:
                    print(f'Query response to {query} invalid')
                    continue
                try:
                    response_tested = whois_query_irrd(self.host_tested, self.port_tested, query)
                except WhoisQueryError as wqe:
                    error_tested = str(wqe)
                except ValueError:
                    print(f'Query response to {query} invalid')
                    continue

            else:  # RIPE style query
                try:
                    response_reference = whois_query(self.host_reference, self.port_reference, query)
                except ConnectionError as ce:
                    error_reference = str(ce)
                response_tested = whois_query(self.host_tested, self.port_tested, query)

            # If both produce error messages, don't compare them
            both_error = error_reference and error_tested
            both_comment = (response_reference and response_tested and
                            response_reference.strip() and response_tested.strip() and
                            response_reference.strip()[0] == '%' and response_tested.strip()[0] == '%')
            if both_error or both_comment:
                self.queries_both_error += 1
                continue

            try:
                cleaned_reference = self.clean(query, response_reference)
            except ValueError as ve:
                print(f'Invalid reference response to query {query.strip()}: {response_reference}: {ve}')
                continue

            try:
                cleaned_tested = self.clean(query, response_tested)
            except ValueError as ve:
                print(f'Invalid tested response to query {query.strip()}: {response_tested}: {ve}')
                continue

            if cleaned_reference != cleaned_tested:
                self.queries_different += 1
                self.write_inconsistency_report(query, cleaned_reference, cleaned_tested)

        print(f'Ran {self.queries_run} objects, {self.queries_different} had different results, '
              f'{self.queries_both_error} produced errors on both instances, '
              f'{self.queries_invalid} invalid queries were skipped, '
              f'{self.queries_mirror} NRTM queries were skipped')

    def clean(self, query: str, response: Optional[str]) -> Optional[str]:
        """Clean the query response, so that the text can be compared."""
        if not response:
            return response
        irr_query = query[:2].lower()
        response = response.strip().lower()

        cleaned_result_list = None
        if irr_query in SSP_QUERIES or (irr_query == '!r' and query.lower().strip().endswith(',o')):
            cleaned_result_list = response.split(' ')
        if irr_query in ['!6', '!g'] and cleaned_result_list:
            cleaned_result_list = [str(IP(ip)) for ip in cleaned_result_list]
        if cleaned_result_list:
            return ' '.join(sorted(list(set(cleaned_result_list))))
        else:
            new_responses = []
            for paragraph in split_paragraphs_rpsl(response):
                rpsl_obj = rpsl_object_from_text(paragraph.strip(), strict_validation=False)
                new_responses.append(rpsl_obj)

            new_responses.sort(key=lambda i: i.parsed_data.get('source', '') + i.rpsl_object_class + i.pk())
            texts = [r.render_rpsl_text() for r in new_responses]
            return '\n'.join(OrderedSet(texts))

    def write_inconsistency_report(self, query, cleaned_reference, cleaned_tested):
        """Write a report to disk with details of the query response inconsistency."""
        report = open(f'qout/QR {query.strip().replace("/", "S")[:30]}', 'w')
        diff_str = self.render_diff(query, cleaned_reference, cleaned_tested)
        report.write(query.strip() + '\n')
        report.write('\n=================================================================\n')
        if diff_str:
            report.write(f'~~~~~~~~~[ diff clean ref->tst ]~~~~~~~~~\n')
            report.write(diff_str + '\n')
        report.write(f'~~~~~~~~~[ clean ref {self.host_reference}:{self.port_reference} ]~~~~~~~~~\n')
        report.write(str(cleaned_reference) + '\n')
        report.write(f'~~~~~~~~~[ clean tst {self.host_tested}:{self.port_tested} ]~~~~~~~~~\n')
        report.write(str(cleaned_tested) + '\n')
        report.write('\n=================================================================\n')
        report.close()

    def render_diff(self, query: str, cleaned_reference: str, cleaned_tested: str) -> Optional[str]:
        """Produce a diff between the results, either by line or with queries like !i, by element returned."""
        if not cleaned_reference or not cleaned_tested:
            return None

        irr_query = query[:2].lower()
        if irr_query in SSP_QUERIES or (irr_query == '!r' and query.lower().strip().endswith(',o')):
            diff_input_reference = list(cleaned_reference.split(' '))
            diff_input_tested = list(cleaned_tested.split(' '))
        else:
            diff_input_reference = list(splitline_unicodesafe(cleaned_reference))
            diff_input_tested = list(splitline_unicodesafe(cleaned_tested))
        diff = list(difflib.unified_diff(diff_input_reference, diff_input_tested, lineterm=''))
        diff_str = '\n'.join(diff[2:])  # skip the lines from the diff which would have filenames
        return diff_str


def main():  # pragma: no cover
    description = """Run a list of queries against two IRRD instances, and report significant results."""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('input_file', type=str,
                        help='the name of a file to read containing queries, or - for stdin')
    parser.add_argument('host_reference', type=str,
                        help='host/IP of the reference IRRD server')
    parser.add_argument('port_reference', type=int,
                        help='port for the reference IRRD server')
    parser.add_argument('host_tested', type=str,
                        help='host/IP of the tested IRRD server')
    parser.add_argument('port_tested', type=int,
                        help='port for the tested IRRD server')
    args = parser.parse_args()

    QueryComparison(args.input_file, args.host_reference, args.port_reference, args.host_tested, args.port_tested)


if __name__ == '__main__':
    main()



"""Add synchronised_serial to database status

Revision ID: a7766c144d61
Revises: 4a514ead8fc2
Create Date: 2020-08-29 19:14:41.800696

"""
import sqlalchemy as sa
from alembic import op

from irrd.storage.database_handler import is_serial_synchronised
from irrd.storage.models import RPSLDatabaseStatus

# revision identifiers, used by Alembic.
revision = 'a7766c144d61'
down_revision = '4a514ead8fc2'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('database_status', sa.Column('synchronised_serials', sa.Boolean(), nullable=False, server_default='f'))

    connection = op.get_bind()
    t_db_status = RPSLDatabaseStatus.__table__

    for row in connection.execute(t_db_status.select()):
        synchronised_serials = is_serial_synchronised(None, row['source'], settings_only=True)
        connection.execute(t_db_status.update().where(
            t_db_status.c.source == row['source']
        ).values(
            synchronised_serials=synchronised_serials,
        ))


def downgrade():
    op.drop_column('database_status', 'synchronised_serials')



from collections import defaultdict

import ujson

import logging
from IPy import IP, IPSet
from typing import List, Optional, Dict, Set

from irrd.conf import RPKI_IRR_PSEUDO_SOURCE, get_setting
from irrd.rpki.status import RPKIStatus
from irrd.rpsl.parser import RPSLObject, RPSL_ATTRIBUTE_TEXT_WIDTH
from irrd.rpsl.rpsl_objects import RPSL_ROUTE_OBJECT_CLASS_FOR_IP_VERSION
from irrd.scopefilter.validators import ScopeFilterValidator
from irrd.storage.database_handler import DatabaseHandler
from irrd.storage.models import JournalEntryOrigin
from irrd.utils.validators import parse_as_number

SLURM_TRUST_ANCHOR = 'SLURM file'

logger = logging.getLogger(__name__)


class ROAParserException(Exception):
    pass


class ROADataImporter:
    """
    Importer for ROAs.

    Loads all ROAs from the JSON data in the rpki_json_str parameter.
    If a slurm_json_str is provided, it is used to filter/amend the ROAs.

    Expects all existing ROA and pseudo-IRR objects to be deleted
    already, e.g. with:
        database_handler.delete_all_roa_objects()
        database_handler.delete_all_rpsl_objects_with_journal(RPKI_IRR_PSEUDO_SOURCE)
    """
    def __init__(self, rpki_json_str: str, slurm_json_str: Optional[str],
                 database_handler: DatabaseHandler):
        self.roa_objs: List[ROA] = []
        self._filtered_asns: Set[int] = set()
        self._filtered_prefixes: IPSet = IPSet()
        self._filtered_combined: Dict[int, IPSet] = defaultdict(IPSet)

        self._load_roa_dicts(rpki_json_str)
        if slurm_json_str:
            self._load_slurm(slurm_json_str)

        scopefilter_validator = ScopeFilterValidator()

        for roa_dict in self._roa_dicts:
            try:
                _, asn = parse_as_number(roa_dict['asn'], permit_plain=True)
                prefix = IP(roa_dict['prefix'])
                ta = roa_dict['ta']
                if ta != SLURM_TRUST_ANCHOR:
                    if asn in self._filtered_asns:
                        continue
                    if any([prefix in self._filtered_prefixes]):
                        continue
                    if any([prefix in self._filtered_combined.get(asn, [])]):
                        continue

                roa_obj = ROA(prefix, asn, roa_dict['maxLength'], ta)
            except KeyError as ke:
                msg = f'Unable to parse ROA record: missing key {ke} -- full record: {roa_dict}'
                logger.error(msg)
                raise ROAParserException(msg)
            except ValueError as ve:
                msg = f'Invalid value in ROA or SLURM: {ve}'
                logger.error(msg)
                raise ROAParserException(msg)

            roa_obj.save(database_handler, scopefilter_validator)
            self.roa_objs.append(roa_obj)

    def _load_roa_dicts(self, rpki_json_str: str) -> None:
        """Load the ROAs from the JSON string into self._roa_dicts"""
        try:
            self._roa_dicts = ujson.loads(rpki_json_str)['roas']
        except ValueError as error:
            msg = f'Unable to parse ROA input: invalid JSON: {error}'
            logger.error(msg)
            raise ROAParserException(msg)
        except KeyError:
            msg = 'Unable to parse ROA input: root key "roas" not found'
            logger.error(msg)
            raise ROAParserException(msg)

    def _load_slurm(self, slurm_json_str: str):
        """
        Load and apply the SLURM in the provided JSON string.

        Prefix filters are loaded into three instance variables:
        - self._filtered_asns: a set of ASNs (in "ASxxx" format) that
          should be filtered, i.e. any ROA with this asn is discarded
        - self._filtered_prefixes: a set of prefixes (as IP instances)
          that should be filtered, i.e. any ROA matching or more specific
          of a prefix in this set should be discarded
        - self._filtered_combined: a dict with ASNs as keys, sets of
          prefixes as values, that should be filered. Any ROA matching
          that ASN and having a prefix that matches or is more specific,
          should be discarded

        Prefix assertions are directly loaded into self._roa_dicts.
        This must be called after _load_roa_dicts()
        """
        slurm = ujson.loads(slurm_json_str)
        version = slurm.get('slurmVersion')
        if version != 1:
            msg = f'SLURM data has invalid version: {version}'
            logger.error(msg)
            raise ROAParserException(msg)

        filters = slurm.get('validationOutputFilters', {}).get('prefixFilters', [])
        for item in filters:
            if 'asn' in item and 'prefix' not in item:
                self._filtered_asns.add(int(item['asn']))
            if 'asn' not in item and 'prefix' in item:
                self._filtered_prefixes.add(IP(item['prefix']))
            if 'asn' in item and 'prefix' in item:
                self._filtered_combined[int(item['asn'])].add(IP(item['prefix']))

        assertions = slurm.get('locallyAddedAssertions', {}).get('prefixAssertions', [])
        for assertion in assertions:
            max_length = assertion.get('maxPrefixLength')
            if max_length is None:
                max_length = IP(assertion['prefix']).prefixlen()
            self._roa_dicts.append({
                'asn': 'AS' + str(assertion['asn']),
                'prefix': assertion['prefix'],
                'maxLength': max_length,
                'ta': SLURM_TRUST_ANCHOR,
            })


class ROA:
    """
    Representation of a ROA.

    This is used when (re-)importing all ROAs, to save the data to the DB,
    and by the BulkRouteROAValidator when validating all existing routes.
    """
    def __init__(self, prefix: IP, asn: int, max_length: str, trust_anchor: str):
        try:
            self.prefix = prefix
            self.prefix_str = str(prefix)
            self.asn = asn
            self.max_length = int(max_length)
            self.trust_anchor = trust_anchor
        except ValueError as ve:
            msg = f'Invalid value in ROA: {ve}'
            logger.error(msg)
            raise ROAParserException(msg)

        if self.max_length < self.prefix.prefixlen():
            msg = f'Invalid ROA: prefix size {self.prefix.prefixlen()} is smaller than max length {max_length} in ' \
                  f'ROA for {self.prefix} / AS{self.asn}'
            logger.error(msg)
            raise ROAParserException(msg)

    def save(self, database_handler: DatabaseHandler, scopefilter_validator: ScopeFilterValidator):
        """
        Save the ROA object to the DB, create a pseudo-IRR object, and save that too.
        """
        database_handler.insert_roa_object(
            ip_version=self.prefix.version(),
            prefix_str=self.prefix_str,
            asn=self.asn,
            max_length=self.max_length,
            trust_anchor=self.trust_anchor,
        )
        self._rpsl_object = RPSLObjectFromROA(
            prefix=self.prefix,
            prefix_str=self.prefix_str,
            asn=self.asn,
            max_length=self.max_length,
            trust_anchor=self.trust_anchor,
            scopefilter_validator=scopefilter_validator,
        )
        database_handler.upsert_rpsl_object(self._rpsl_object, JournalEntryOrigin.pseudo_irr,
                                            rpsl_guaranteed_no_existing=True)


class RPSLObjectFromROA(RPSLObject):
    """
    This is an RPSLObject compatible class that represents
    an RPKI pseudo-IRR object. It overrides the API in
    relevant parts.
    """
    # noinspection PyMissingConstructor
    def __init__(self, prefix: IP, prefix_str: str, asn: int, max_length: int, trust_anchor: str,
                 scopefilter_validator: ScopeFilterValidator):
        self.prefix = prefix
        self.prefix_str = prefix_str
        self.asn = asn
        self.max_length = max_length
        self.trust_anchor = trust_anchor

        self.rpsl_object_class = RPSL_ROUTE_OBJECT_CLASS_FOR_IP_VERSION[self.prefix.version()]
        self.ip_first = self.prefix.net()
        self.ip_last = self.prefix.broadcast()
        self.prefix_length = self.prefix.prefixlen()
        self.asn_first = asn
        self.asn_last = asn
        self.rpki_status = RPKIStatus.valid
        self.parsed_data = {
            self.rpsl_object_class: self.prefix_str,
            'origin': 'AS' + str(self.asn),
            'source': RPKI_IRR_PSEUDO_SOURCE,
        }
        self.scopefilter_status, _ = scopefilter_validator.validate_rpsl_object(self)

    def source(self):
        return RPKI_IRR_PSEUDO_SOURCE

    def pk(self):
        return f'{self.prefix_str}AS{self.asn}/ML{self.max_length}'

    def render_rpsl_text(self, last_modified=None):
        object_class_display = f'{self.rpsl_object_class}:'.ljust(RPSL_ATTRIBUTE_TEXT_WIDTH)
        remarks_fill = RPSL_ATTRIBUTE_TEXT_WIDTH * ' '
        remarks = get_setting('rpki.pseudo_irr_remarks').replace('\n', '\n' + remarks_fill).strip()
        remarks = remarks.format(asn=self.asn, prefix=self.prefix_str)
        rpsl_object_text = f"""
{object_class_display}{self.prefix_str}
descr:          RPKI ROA for {self.prefix_str} / AS{self.asn}
remarks:        {remarks}
max-length:     {self.max_length}
origin:         AS{self.asn}
source:         {RPKI_IRR_PSEUDO_SOURCE}  # Trust Anchor: {self.trust_anchor}
""".strip() + '\n'
        return rpsl_object_text
